from __future__ import annotations

import operator
from operator import methodcaller

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from packaging.version import parse as vparse
from pytest import param

import ibis
import ibis.expr.datatypes as dt

dask = pytest.importorskip("dask")
da = pytest.importorskip("dask.array")
dd = pytest.importorskip("dask.dataframe")

from dask.dataframe.utils import tm  # noqa: E402


def test_table_column(t, pandas_df):
    expr = t.plain_int64
    result = expr.execute()
    expected = pandas_df.plain_int64
    tm.assert_series_equal(result, expected)


def test_literal(client):
    assert client.execute(ibis.literal(1)) == 1


def test_selection(t, df):
    expr = t[((t.plain_strings == "a") | (t.plain_int64 == 3)) & (t.dup_strings == "d")]
    result = expr.compile()
    expected = df[
        ((df.plain_strings == "a") | (df.plain_int64 == 3)) & (df.dup_strings == "d")
    ]
    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


def test_mutate(t, df):
    expr = t.mutate(x=t.plain_int64 + 1, y=t.plain_int64 * 2)
    result = expr.compile()
    expected = df.assign(x=df.plain_int64 + 1, y=df.plain_int64 * 2)
    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.xfail(reason="TODO - windowing - #2553")
def test_project_scope_does_not_override(t, df):
    col = t.plain_int64
    expr = t[
        [
            col.name("new_col"),
            col.sum().over(ibis.window(group_by="dup_strings")).name("grouped"),
        ]
    ]
    result = expr.compile()
    expected = dd.concat(
        [
            df[["plain_int64", "dup_strings"]].rename(
                columns={"plain_int64": "new_col"}
            ),
            df.groupby("dup_strings")
            .plain_int64.transform("sum")
            .reset_index(drop=True)
            .rename("grouped"),
        ],
        axis=1,
    )[["new_col", "grouped"]]
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.parametrize(
    "where",
    [
        param(lambda _: None, id="none"),
        param(lambda t: t.dup_strings == "d", id="simple"),
        param(lambda t: (t.dup_strings == "d") | (t.plain_int64 < 100), id="complex"),
    ],
)
@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        param(methodcaller("abs"), np.abs, id="abs"),
        param(methodcaller("ceil"), np.ceil, id="ceil"),
        param(methodcaller("exp"), np.exp, id="exp"),
        param(methodcaller("floor"), np.floor, id="floor"),
        param(methodcaller("ln"), np.log, id="log"),
        param(methodcaller("log10"), np.log10, id="log10"),
        param(methodcaller("log", 2), lambda x: np.log(x) / np.log(2), id="logb"),
        param(methodcaller("log2"), np.log2, id="log2"),
        param(
            methodcaller("round", 0), lambda x: x.round(0).astype("int64"), id="round0"
        ),
        param(methodcaller("round", -2), methodcaller("round", -2), id="roundm2"),
        param(methodcaller("round", 2), methodcaller("round", 2), id="round2"),
        param(methodcaller("round"), lambda x: x.round().astype("int64"), id="round"),
        param(methodcaller("sign"), np.sign, id="sign"),
        param(methodcaller("sqrt"), np.sqrt, id="sqrt"),
    ],
)
def test_aggregation_group_by(t, pandas_df, where, ibis_func, pandas_func):
    ibis_where = where(t)
    expr = t.group_by(t.dup_strings).aggregate(
        avg_plain_int64=t.plain_int64.mean(where=ibis_where),
        sum_plain_float64=t.plain_float64.sum(where=ibis_where),
        mean_float64_positive=ibis_func(t.float64_positive).mean(where=ibis_where),
        neg_mean_int64_with_zeros=(-t.int64_with_zeros).mean(where=ibis_where),
        nunique_dup_ints=t.dup_ints.nunique(),
    )
    result = expr.execute()

    df = pandas_df
    pandas_where = where(df)
    mask = slice(None) if pandas_where is None else pandas_where
    expected = (
        df.groupby("dup_strings")
        .agg(
            {
                "plain_int64": lambda x, mask=mask: x[mask].mean(),
                "plain_float64": lambda x, mask=mask: x[mask].sum(),
                "dup_ints": "nunique",
                "float64_positive": (
                    lambda x, mask=mask, func=pandas_func: func(x[mask]).mean()
                ),
                "int64_with_zeros": lambda x, mask=mask: (-x[mask]).mean(),
            }
        )
        .reset_index()
        .rename(
            columns={
                "plain_int64": "avg_plain_int64",
                "plain_float64": "sum_plain_float64",
                "dup_ints": "nunique_dup_ints",
                "float64_positive": "mean_float64_positive",
                "int64_with_zeros": "neg_mean_int64_with_zeros",
            }
        )
    )
    lhs = result[expected.columns]
    rhs = expected
    tm.assert_frame_equal(lhs, rhs)


def test_aggregation_without_group_by(t, df):
    expr = t.aggregate(
        avg_plain_int64=t.plain_int64.mean(),
        sum_plain_float64=t.plain_float64.sum(),
    )
    result = expr.compile()[["avg_plain_int64", "sum_plain_float64"]]
    new_names = {
        "plain_float64": "sum_plain_float64",
        "plain_int64": "avg_plain_int64",
    }
    pandas_df = df.compute().reset_index(drop=True)
    expected = (
        pd.Series(
            [
                pandas_df["plain_int64"].mean(),
                pandas_df["plain_float64"].sum(),
            ],
            index=["plain_int64", "plain_float64"],
        )
        .to_frame()
        .T.rename(columns=new_names)
    )
    lhs = result[expected.columns].compute().reset_index(drop=True)
    tm.assert_frame_equal(lhs, expected)


def test_group_by_with_having(t, df):
    expr = (
        t.group_by(t.dup_strings)
        .having(t.plain_float64.sum() == 5)
        .aggregate(avg_a=t.plain_int64.mean(), sum_c=t.plain_float64.sum())
    )
    result = expr.compile()

    expected = (
        df.groupby("dup_strings")
        .agg({"plain_int64": "mean", "plain_float64": "sum"})
        .reset_index()
        .rename(columns={"plain_int64": "avg_a", "plain_float64": "sum_c"})
    )
    expected = expected.loc[expected.sum_c == 5, ["avg_a", "sum_c"]]

    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


def test_group_by_rename_key(t, df):
    expr = t.group_by(t.dup_strings.name("foo")).aggregate(
        dup_string_count=t.dup_strings.count()
    )

    assert "foo" in expr.schema()
    result = expr.compile()
    assert "foo" in result.columns

    expected = (
        df.groupby("dup_strings")
        .dup_strings.count()
        .rename("dup_string_count")
        .reset_index()
        .rename(columns={"dup_strings": "foo"})
    )
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.parametrize("reduction", ["mean", "sum", "count", "std", "var"])
@pytest.mark.parametrize(
    "where",
    [
        lambda t: (t.plain_strings == "a") | (t.plain_strings == "c"),
        lambda t: (t.dup_strings == "d")
        & ((t.plain_int64 == 1) | (t.plain_int64 == 3)),
        lambda t: None,
    ],
)
def test_reduction(t, pandas_df, reduction, where):
    func = getattr(t.plain_int64, reduction)
    mask = where(t)
    expr = func(where=mask)
    result = expr.execute()

    df_mask = where(pandas_df)
    expected_func = getattr(
        pandas_df.loc[df_mask if df_mask is not None else slice(None), "plain_int64"],
        reduction,
    )
    expected = expected_func()
    assert result == expected


@pytest.mark.parametrize(
    "where",
    [
        lambda t: (t.plain_strings == "a") | (t.plain_strings == "c"),
        lambda t: None,
    ],
)
def test_grouped_reduction(t, df, where):
    ibis_where = where(t)
    expr = t.group_by(t.dup_strings).aggregate(
        nunique_dup_ints=t.dup_ints.nunique(),
        sum_plain_int64=t.plain_int64.sum(where=ibis_where),
        mean_plain_int64=t.plain_int64.mean(where=ibis_where),
        count_plain_int64=t.plain_int64.count(where=ibis_where),
        std_plain_int64=t.plain_int64.std(where=ibis_where),
        var_plain_int64=t.plain_int64.var(where=ibis_where),
        nunique_plain_int64=t.plain_int64.nunique(where=ibis_where),
    )
    result = expr.compile()

    df_mask = where(df.compute())
    mask = slice(None) if df_mask is None else df_mask

    expected = (
        df.compute()
        .groupby("dup_strings")
        .agg(
            {
                "dup_ints": "nunique",
                "plain_int64": [
                    lambda x, mask=mask: x[mask].sum(),
                    lambda x, mask=mask: x[mask].mean(),
                    lambda x, mask=mask: x[mask].count(),
                    lambda x, mask=mask: x[mask].std(),
                    lambda x, mask=mask: x[mask].var(),
                    lambda x, mask=mask: x[mask].nunique(),
                ],
            }
        )
        .reset_index()
    )
    result = result.compute()

    assert len(result.columns) == len(expected.columns)

    expected.columns = [
        "dup_strings",
        "nunique_dup_ints",
        "sum_plain_int64",
        "mean_plain_int64",
        "count_plain_int64",
        "std_plain_int64",
        "var_plain_int64",
        "nunique_plain_int64",
    ]
    # guarantee ordering
    result = result[expected.columns]
    # dask and pandas differ slightly in how they treat groups with no entry
    # we're not testing that so fillna here.
    result = result.fillna(0.0)
    expected = expected.fillna(0.0)

    # match the dtypes
    if df_mask is None:
        expected["mean_plain_int64"] = expected.mean_plain_int64.astype("float64")
    else:
        expected["sum_plain_int64"] = expected.sum_plain_int64.astype("int64")
        expected["count_plain_int64"] = expected.count_plain_int64.astype("int64")
        expected["nunique_plain_int64"] = expected.nunique_plain_int64.astype("int64")

    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "reduction",
    [
        lambda x: x.any(),
        lambda x: x.all(),
        lambda x: ~(x.any()),
        lambda x: ~(x.all()),
    ],
)
def test_boolean_aggregation(t, pandas_df, reduction):
    expr = reduction(t.plain_int64 == 1)
    result = expr.execute()
    expected = reduction(pandas_df.plain_int64 == 1)
    assert result == expected


@pytest.mark.parametrize("column", ["float64_with_zeros", "int64_with_zeros"])
def test_nullif_zero(t, pandas_df, column):
    expr = t[column].nullif(0)
    result = expr.execute()
    expected = pandas_df[column].replace(0, np.nan)
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


@pytest.mark.parametrize(
    ("left", "right", "expected", "compare"),
    [
        param(
            lambda t: ibis.literal(1),
            lambda t: ibis.literal(1),
            lambda df: np.nan,
            np.testing.assert_array_equal,  # treats NaNs as equal
            id="literal_literal_equal",
        ),
        param(
            lambda t: ibis.literal(1),
            lambda t: ibis.literal(2),
            lambda df: 1,
            np.testing.assert_equal,
            id="literal_literal_not_equal",
        ),
        param(
            lambda t: t.dup_strings,
            lambda t: ibis.literal("a"),
            lambda df: df.dup_strings.where(df.dup_strings != "a"),
            tm.assert_series_equal,
            id="series_literal",
        ),
        param(
            lambda t: t.dup_strings,
            lambda t: t.dup_strings,
            lambda df: df.dup_strings.where(df.dup_strings != df.dup_strings),
            tm.assert_series_equal,
            id="series_series",
        ),
        param(
            lambda t: ibis.literal("a"),
            lambda t: t.dup_strings,
            lambda _: pd.Series(["a", np.nan, "a"], name="dup_strings"),
            tm.assert_series_equal,
            id="literal_series",
        ),
    ],
)
def test_nullif(t, con, pandas_df, left, right, expected, compare):
    expr = left(t).nullif(right(t))
    result = con.execute(expr.name("dup_strings"))
    compare(result, expected(pandas_df))


def test_nullif_inf(con):
    df = pd.DataFrame({"a": [np.inf, 3.14, -np.inf, 42.0]})
    t = ibis.memtable(df)
    expr = t.a.nullif(np.inf).nullif(-np.inf)
    result = con.execute(expr)
    expected = pd.Series([np.nan, 3.14, np.nan, 42.0], name="a")
    tm.assert_series_equal(result, expected, check_names=False)


def test_group_concat(t, df):
    expr = (
        t[t.dup_ints == 1]
        .group_by(t.dup_strings)
        .aggregate(foo=t.dup_ints.group_concat(","))
    )
    result = expr.compile()
    expected = (
        df[df.dup_ints == 1]
        .groupby("dup_strings")
        .apply(lambda df: ",".join(df.dup_ints.astype(str)))
        .reset_index()
        .rename(columns={0: "foo"})
    )

    left = (
        result[expected.columns]
        .compute()
        .sort_values("dup_strings")
        .reset_index(drop=True)
    )
    right = expected.compute().sort_values("dup_strings").reset_index(drop=True)
    tm.assert_frame_equal(left, right)


@pytest.mark.parametrize("offset", [0, 2])
def test_frame_limit(t, df, offset):
    n = 5
    df_expr = t.limit(n, offset=offset)
    result = df_expr.compile()
    expected = df.loc[offset : offset + n].reset_index(drop=True)
    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.xfail(raises=AttributeError, reason="TableColumn does not implement limit")
@pytest.mark.parametrize("offset", [0, 2])
def test_series_limit(t, df, offset):
    n = 5
    s_expr = t.plain_int64.limit(n, offset=offset)
    result = s_expr.compile()
    tm.assert_series_equal(
        result, df.plain_int64.iloc[offset : offset + n], check_index=False
    )


@pytest.mark.xfail(
    condition=vparse(dask.__version__) < vparse("2024.2.0"),
    reason="not implemented until 2024.2.0",
)
def test_complex_order_by(t, df):
    expr = t.order_by([ibis.desc(t.plain_int64 * t.plain_float64), t.plain_float64])
    result = expr.compile()
    expected = (
        df.assign(foo=df.plain_int64 * df.plain_float64)
        .sort_values(["foo", "plain_float64"], ascending=[False, True])
        .drop(["foo"], axis=1)
        .reset_index(drop=True)
    )

    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


def test_count_distinct(t, pandas_df):
    expr = t.dup_strings.nunique()
    result = expr.execute()
    expected = pandas_df.dup_strings.nunique()
    assert result == expected


def test_value_counts(t, df):
    expr = t.dup_strings.value_counts()
    result = expr.compile()
    expected = (
        df.compute()
        .dup_strings.value_counts()
        .rename("dup_strings")
        .reset_index(name="dup_strings_count")
        .rename(columns={"index": "dup_strings"})
        .sort_values(["dup_strings"])
        .reset_index(drop=True)
    )
    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True), expected
    )


def test_table_count(t, df):
    expr = t.count()
    result = expr.execute()
    expected = len(df)
    assert result == expected


def test_weighted_average(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        avg=(t.plain_float64 * t.plain_int64).sum() / t.plain_int64.sum()
    )
    result = expr.compile()
    expected = (
        df.groupby("dup_strings")
        .apply(
            lambda df: (df.plain_int64 * df.plain_float64).sum() / df.plain_int64.sum()
        )
        .reset_index()
        .rename(columns={0: "avg"})
    )
    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


def test_group_by_multiple_keys(t, df):
    expr = t.group_by([t.dup_strings, t.dup_ints]).aggregate(
        avg_plain_float64=t.plain_float64.mean()
    )
    result = expr.compile()
    expected = (
        df.groupby(["dup_strings", "dup_ints"])
        .agg({"plain_float64": "mean"})
        .reset_index()
        .rename(columns={"plain_float64": "avg_plain_float64"})
    )
    tm.assert_frame_equal(
        result[expected.columns]
        .compute()
        .sort_values(["dup_strings", "dup_ints"])
        .reset_index(drop=True),
        expected.compute()
        .sort_values(["dup_strings", "dup_ints"])
        .reset_index(drop=True),
    )


def test_mutate_after_group_by(t, df):
    gb = t.group_by(t.dup_strings).aggregate(avg_plain_float64=t.plain_float64.mean())
    expr = gb.mutate(x=gb.avg_plain_float64)
    result = expr.compile()
    expected = (
        df.groupby("dup_strings")
        .agg({"plain_float64": "mean"})
        .reset_index()
        .rename(columns={"plain_float64": "avg_plain_float64"})
    )
    expected = expected.assign(x=expected.avg_plain_float64)
    tm.assert_frame_equal(
        result[expected.columns]
        .compute()
        .sort_values("dup_strings")
        .reset_index(drop=True),
        expected.compute().sort_values("dup_strings").reset_index(drop=True),
    )


def test_group_by_with_unnamed_arithmetic(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        naive_variance=((t.plain_float64**2).sum() - t.plain_float64.mean() ** 2)
        / t.plain_float64.count()
    )
    result = expr.compile()
    expected = (
        df.compute()
        .groupby("dup_strings")
        .agg({"plain_float64": lambda x: ((x**2).sum() - x.mean() ** 2) / x.count()})
        .reset_index()
        .rename(columns={"plain_float64": "naive_variance"})
    )
    tm.assert_frame_equal(
        result[expected.columns].compute().reset_index(drop=True), expected
    )


def test_isnull(t, pandas_df):
    expr = t.strings_with_nulls.isnull()
    result = expr.execute()
    expected = pandas_df.strings_with_nulls.isnull()
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


def test_notnull(t, pandas_df):
    expr = t.strings_with_nulls.notnull()
    result = expr.execute()
    expected = pandas_df.strings_with_nulls.notnull()
    tm.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("raw_value", [0.0, 1.0])
def test_scalar_parameter(t, pandas_df, raw_value):
    value = ibis.param(dt.double)
    expr = t.float64_with_zeros == value
    result = expr.execute(params={value: raw_value})
    expected = pandas_df.float64_with_zeros == raw_value
    tm.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("elements", [[1], (1,), {1}, frozenset({1})])
def test_isin(t, pandas_df, elements):
    expr = t.plain_float64.isin(elements)
    expected = pandas_df.plain_float64.isin(elements)
    result = expr.execute()
    tm.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("elements", [[1], (1,), {1}, frozenset({1})])
def test_notin(t, pandas_df, elements):
    expr = t.plain_float64.notin(elements)
    expected = ~pandas_df.plain_float64.isin(elements)
    result = expr.execute()
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


def test_cast_on_group_by(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        casted=(t.float64_with_zeros == 0).cast("int64").sum()
    )

    result = expr.compile()
    expected = (
        df.groupby("dup_strings")
        .float64_with_zeros.apply(lambda s: (s == 0).astype("int64").sum())
        .reset_index()
        .rename(columns={"float64_with_zeros": "casted"})
    )
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.parametrize(
    "op",
    [
        operator.add,
        operator.mul,
        operator.sub,
        operator.truediv,
        operator.floordiv,
        operator.mod,
        operator.pow,
    ],
    ids=operator.attrgetter("__name__"),
)
@pytest.mark.parametrize("args", [lambda c: (1.0, c), lambda c: (c, 1.0)])
def test_left_binary_op(t, pandas_df, op, args):
    expr = op(*args(t.float64_with_zeros))
    result = expr.execute()
    expected = op(*args(pandas_df.float64_with_zeros)).astype(result.dtype)
    tm.assert_series_equal(result, expected, check_index=False, check_names=False)


@pytest.mark.parametrize(
    "op",
    [
        operator.add,
        operator.mul,
        operator.sub,
        operator.truediv,
        operator.floordiv,
        operator.mod,
        operator.pow,
    ],
    ids=operator.attrgetter("__name__"),
)
@pytest.mark.parametrize("argfunc", [lambda c: (1.0, c), lambda c: (c, 1.0)])
def test_left_binary_op_gb(t, pandas_df, op, argfunc):
    expr = t.group_by("dup_strings").aggregate(
        foo=op(*argfunc(t.float64_with_zeros)).sum()
    )
    result = expr.execute()
    expected = (
        pandas_df.groupby("dup_strings")
        .float64_with_zeros.apply(lambda s: op(*argfunc(s)).sum())
        .reset_index()
        .rename(columns={"float64_with_zeros": "foo"})
    )
    expected["foo"] = expected["foo"].astype(result["foo"].dtype)
    tm.assert_frame_equal(result, expected, check_names=False)


@pytest.mark.parametrize(
    "left_f",
    [
        param(lambda e: e - 1, id="sub"),
        param(lambda _: 0.0, id="zero"),
        param(lambda _: None, id="none"),
    ],
)
@pytest.mark.parametrize(
    "right_f",
    [
        param(lambda e: e + 1, id="add"),
        param(lambda _: 1.0, id="one"),
        param(lambda _: None, id="none"),
    ],
)
def test_ifelse_series(t, pandas_df, left_f, right_f):
    col_expr = t["plain_int64"]
    result = ibis.ifelse(
        col_expr > col_expr.mean(), left_f(col_expr), right_f(col_expr)
    ).execute()

    series = pandas_df["plain_int64"]
    cond = series > series.mean()
    left = left_f(series)
    if not isinstance(left, pd.Series):
        left = pd.Series(np.repeat(left, len(cond)), name=cond.name)
    expected = left.where(cond, right_f(series))

    tm.assert_series_equal(
        result.astype(object).fillna(pd.NA),
        expected.astype(object).fillna(pd.NA),
        check_dtype=False,
        check_names=False,
    )


@pytest.mark.parametrize(
    ("cond", "expected_func"),
    [
        param(True, lambda df: df["plain_int64"].astype("float64"), id="true"),
        param(False, lambda df: pd.Series(np.repeat(3.0, len(df))), id="false"),
    ],
)
def test_ifelse_scalar(t, pandas_df, cond, expected_func):
    expr = ibis.ifelse(cond, t["plain_int64"], 3.0)
    result = expr.execute()
    expected = expected_func(pandas_df)
    tm.assert_series_equal(result, expected, check_names=False)


def test_ifelse_long(batting, batting_pandas_df):
    col_expr = batting["AB"]
    result = ibis.ifelse(col_expr > col_expr.mean(), col_expr, 0.0).execute()

    series = batting_pandas_df["AB"]
    expected = series.where(series > series.mean(), other=0.0).astype("float64")

    tm.assert_series_equal(result, expected, check_names=False)


def test_round(t, pandas_df):
    precision = 2
    mult = 3.33333
    result = (t.count() * mult).round(precision).execute()
    expected = np.around(len(pandas_df) * mult, precision)
    npt.assert_almost_equal(result, expected, decimal=precision)


def test_quantile_group_by(batting, batting_pandas_df):
    def q_fun(x, quantile):
        res = x.quantile(quantile).tolist()
        return [res for _ in range(len(x))]

    frac = 0.2
    result = (
        batting.group_by("teamID")
        .mutate(res=lambda x: x.RBI.quantile([frac, 1 - frac]))
        .res.execute()
    )
    expected = (
        batting_pandas_df.groupby("teamID")
        .RBI.transform(q_fun, quantile=[frac, 1 - frac])
        .rename("res")
    )
    tm.assert_series_equal(result, expected, check_index=False)


def test_searched_case_scalar(client):
    expr = ibis.case().when(True, 1).when(False, 2).end()
    result = client.execute(expr)
    expected = np.int8(1)
    assert result == expected


def test_searched_case_column(batting, batting_pandas_df):
    t = batting
    df = batting_pandas_df
    expr = (
        ibis.case()
        .when(t.RBI < 5, "really bad team")
        .when(t.teamID == "PH1", "ph1 team")
        .else_(t.teamID)
        .end()
    )
    result = expr.execute()
    expected = pd.Series(
        np.select(
            [df.RBI < 5, df.teamID == "PH1"],
            ["really bad team", "ph1 team"],
            df.teamID,
        )
    )
    tm.assert_series_equal(result, expected, check_names=False)


def test_simple_case_scalar(client):
    x = ibis.literal(2)
    expr = x.case().when(2, x - 1).when(3, x + 1).when(4, x + 2).end()
    result = client.execute(expr)
    expected = np.int8(1)
    assert result == expected


def test_simple_case_column(batting, batting_pandas_df):
    t = batting
    df = batting_pandas_df
    expr = (
        t.RBI.case()
        .when(5, "five")
        .when(4, "four")
        .when(3, "three")
        .else_("could be good?")
        .end()
    )
    result = expr.execute()
    expected = pd.Series(
        np.select(
            [df.RBI == 5, df.RBI == 4, df.RBI == 3],
            ["five", "four", "three"],
            "could be good?",
        )
    )
    tm.assert_series_equal(result, expected, check_names=False)


def test_table_distinct(t, df):
    expr = t[["dup_strings"]].distinct()
    result = expr.compile()
    expected = df[["dup_strings"]].drop_duplicates()
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


@pytest.mark.parametrize("distinct", [True, False])
def test_union(client, df1, distinct):
    t = client.table("df1")
    expr = t.union(t, distinct=distinct)
    result = expr.compile()
    expected = df1 if distinct else dd.concat([df1, df1], axis=0, ignore_index=True)

    # match indices because of dask reset_index behavior
    result = result.compute().reset_index(drop=True)
    expected = expected.compute().reset_index(drop=True)

    tm.assert_frame_equal(result, expected)


def test_intersect(client, df1, intersect_df2):
    t1 = client.table("df1")
    t2 = client.table("intersect_df2")
    expr = t1.intersect(t2)
    result = expr.compile()
    expected = df1.merge(intersect_df2, on=list(df1.columns))
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )


def test_difference(client, df1, intersect_df2):
    t1 = client.table("df1")
    t2 = client.table("intersect_df2")
    expr = t1.difference(t2)
    result = expr.compile()
    merged = df1.merge(intersect_df2, on=list(df1.columns), how="outer", indicator=True)
    expected = merged[merged["_merge"] != "both"].drop("_merge", axis=1)

    # force same index
    result = result.compute().reset_index(drop=True)
    expected = expected.compute().reset_index(drop=True)

    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "distinct",
    [
        param(
            True,
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason="dask cannot compute the distinct element of an array column",
            ),
        ),
        False,
    ],
)
def test_union_with_list_types(t, df, distinct):
    expr = t.union(t, distinct=distinct)
    result = expr.compile()
    expected = df if distinct else dd.concat([df, df], axis=0, ignore_index=True)
    tm.assert_frame_equal(
        result.compute().reset_index(drop=True),
        expected.compute().reset_index(drop=True),
    )



from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import NamedTuple

import numpy as np
import pandas.testing as tm
import pyarrow as pa
import pytest

import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from ibis.common.exceptions import IntegrityError
from ibis.common.grounds import Annotable
from ibis.common.patterns import CoercedTo

has_pandas = False
with contextlib.suppress(ImportError):
    import pandas as pd

    has_pandas = True

has_dask = False
with contextlib.suppress(ImportError):
    import dask.dataframe as dd  # noqa: F401

    has_dask = True


def test_whole_schema():
    schema = {
        "cid": "int64",
        "mktsegment": "string",
        "address": """struct<city: string,
                             street: string,
                             street_number: int32,
                             zip: int16>""",
        "phone_numbers": "array<string>",
        "orders": """array<struct<oid: int64,
                                  status: string,
                                  totalprice: decimal(12, 2),
                                  order_date: string,
                                  items: array<struct<iid: int64,
                                                      name: string,
                                                      price: decimal(12, 2),
                                                      discount_perc: decimal(12, 2),
                                                      shipdate: string>>>>
                                                      """,
        "web_visits": """map<string, struct<user_agent: string,
                                            client_ip: string,
                                            visit_date: string,
                                            duration_ms: int32>>""",
        "support_calls": """array<struct<agent_id: int64,
                                         call_date: string,
                                         duration_ms: int64,
                                         issue_resolved: boolean,
                                         agent_comment: string>>""",
    }
    expected = {
        "cid": dt.int64,
        "mktsegment": dt.string,
        "address": dt.Struct(
            {
                "city": dt.string,
                "street": dt.string,
                "street_number": dt.int32,
                "zip": dt.int16,
            }
        ),
        "phone_numbers": dt.Array(dt.string),
        "orders": dt.Array(
            dt.Struct(
                {
                    "oid": dt.int64,
                    "status": dt.string,
                    "totalprice": dt.Decimal(12, 2),
                    "order_date": dt.string,
                    "items": dt.Array(
                        dt.Struct(
                            {
                                "iid": dt.int64,
                                "name": dt.string,
                                "price": dt.Decimal(12, 2),
                                "discount_perc": dt.Decimal(12, 2),
                                "shipdate": dt.string,
                            }
                        )
                    ),
                }
            )
        ),
        "web_visits": dt.Map(
            dt.string,
            dt.Struct(
                {
                    "user_agent": dt.string,
                    "client_ip": dt.string,
                    "visit_date": dt.string,
                    "duration_ms": dt.int32,
                }
            ),
        ),
        "support_calls": dt.Array(
            dt.Struct(
                {
                    "agent_id": dt.int64,
                    "call_date": dt.string,
                    "duration_ms": dt.int64,
                    "issue_resolved": dt.boolean,
                    "agent_comment": dt.string,
                }
            )
        ),
    }
    assert sch.schema(schema) == sch.schema(expected)


def test_schema_from_tuples():
    schema = sch.Schema.from_tuples(
        [
            ("a", "int64"),
            ("b", "string"),
            ("c", "double"),
            ("d", "boolean"),
        ]
    )
    expected = sch.Schema(
        {"a": dt.int64, "b": dt.string, "c": dt.double, "d": dt.boolean}
    )

    assert schema == expected

    # test that duplicate field names are prohibited
    with pytest.raises(IntegrityError):
        sch.Schema.from_tuples([("a", "int64"), ("a", "string")])


def test_schema_subset():
    s1 = sch.schema([("a", dt.int64), ("b", dt.int32), ("c", dt.string)])
    s2 = sch.schema([("a", dt.int64), ("c", dt.string)])

    assert s1 > s2
    assert s2 < s1

    assert s1 >= s2
    assert s2 <= s1


def test_empty_schema():
    s1 = sch.Schema({})
    s2 = sch.schema([])

    assert s1 == s2

    for s in [s1, s2]:
        assert len(s.items()) == 0
        assert repr(s) == "ibis.Schema {\n}"


def test_nullable_output():
    s = sch.schema(
        [
            ("foo", "int64"),
            ("bar", dt.int64(nullable=False)),
            ("baz", "boolean"),
        ]
    )

    sch_str = str(s)
    assert "foo  int64" in sch_str
    assert "foo  !int64" not in sch_str
    assert "bar  !int64" in sch_str
    assert "baz  boolean" in sch_str
    assert "baz  !boolean" not in sch_str


@pytest.fixture
def df():
    return pd.DataFrame({"A": pd.Series([1], dtype="int8"), "b": ["x"]})


def test_apply_to_column_rename(df):
    schema = sch.Schema({"a": "int8", "B": "string"})
    expected = df.rename({"A": "a", "b": "B"}, axis=1)
    with pytest.warns(FutureWarning):
        df = schema.apply_to(df.copy())
    tm.assert_frame_equal(df, expected)


def test_apply_to_column_order(df):
    schema = sch.Schema({"a": "int8", "b": "string"})
    expected = df.rename({"A": "a"}, axis=1)
    with pytest.warns(FutureWarning):
        new_df = schema.apply_to(df.copy())
    tm.assert_frame_equal(new_df, expected)


def test_api_accepts_schema_objects():
    s1 = sch.schema(dict(a="int", b="str"))
    s2 = sch.schema(s1)
    assert s1 == s2


def test_schema_mapping_api():
    s = sch.Schema(
        {
            "a": "map<double, string>",
            "b": "array<map<string, array<int32>>>",
            "c": "array<string>",
            "d": "int8",
        }
    )

    assert s["a"] == dt.Map(dt.double, dt.string)
    assert s["b"] == dt.Array(dt.Map(dt.string, dt.Array(dt.int32)))
    assert s["c"] == dt.Array(dt.string)
    assert s["d"] == dt.int8

    assert "a" in s
    assert "e" not in s
    assert len(s) == 4
    assert tuple(s) == s.names
    assert tuple(s.keys()) == s.names
    assert tuple(s.values()) == s.types
    assert tuple(s.items()) == tuple(zip(s.names, s.types))


class BarSchema:
    a: int
    b: str


class FooSchema:
    a: int
    b: str
    c: float
    d: tuple[str]
    e: list[int]
    f: dict[str, int]
    g: BarSchema
    h: list[BarSchema]
    j: dict[str, BarSchema]


foo_schema = sch.Schema(
    {
        "a": "int64",
        "b": "string",
        "c": "float64",
        "d": "array<string>",
        "e": "array<int64>",
        "f": "map<string, int64>",
        "g": "struct<a: int64, b: string>",
        "h": "array<struct<a: int64, b: string>>",
        "j": "map<string, struct<a: int64, b: string>>",
    }
)


def test_schema_from_annotated_class():
    assert sch.schema(FooSchema) == foo_schema


class NamedBar(NamedTuple):
    a: int
    b: str


class NamedFoo(NamedTuple):
    a: int
    b: str
    c: float
    d: tuple[str]
    e: list[int]
    f: dict[str, int]
    g: NamedBar
    h: list[NamedBar]
    j: dict[str, NamedBar]


def test_schema_from_namedtuple():
    assert sch.schema(NamedFoo) == foo_schema


@dataclass
class DataBar:
    a: int
    b: str


@dataclass
class DataFooBase:
    a: int
    b: str
    c: float
    d: tuple[str]


@dataclass
class DataFoo(DataFooBase):
    e: list[int]
    f: dict[str, int]
    g: DataBar
    h: list[DataBar]
    j: dict[str, DataBar]


def test_schema_from_dataclass():
    assert sch.schema(DataFoo) == foo_schema


class PreferenceA:
    a: dt.int64
    b: dt.Array(dt.int64)


class PreferenceB:
    a: dt.int64
    b: dt.Array[dt.int64]


class PreferenceC:
    a: dt.Int64
    b: dt.Array[dt.Int64]


def test_preferences():
    a = sch.schema(PreferenceA)
    b = sch.schema(PreferenceB)
    c = sch.schema(PreferenceC)
    assert a == b == c


class ObjectWithSchema(Annotable):
    schema: sch.Schema


def test_schema_is_coercible():
    s = sch.Schema({"a": dt.int64, "b": dt.Array(dt.int64)})
    assert CoercedTo(sch.Schema).match(PreferenceA, {}) == s

    o = ObjectWithSchema(schema=PreferenceA)
    assert o.schema == s


def test_schema_set_operations():
    a = sch.Schema({"a": dt.string, "b": dt.int64, "c": dt.float64})
    b = sch.Schema({"a": dt.string, "c": dt.float64, "d": dt.boolean, "e": dt.date})
    c = sch.Schema({"i": dt.int64, "j": dt.float64, "k": dt.string})
    d = sch.Schema({"i": dt.int64, "j": dt.float64, "k": dt.string, "l": dt.boolean})

    assert a & b == sch.Schema({"a": dt.string, "c": dt.float64})
    assert a | b == sch.Schema(
        {"a": dt.string, "b": dt.int64, "c": dt.float64, "d": dt.boolean, "e": dt.date}
    )
    assert a - b == sch.Schema({"b": dt.int64})
    assert b - a == sch.Schema({"d": dt.boolean, "e": dt.date})
    assert a ^ b == sch.Schema({"b": dt.int64, "d": dt.boolean, "e": dt.date})

    assert not a.isdisjoint(b)
    assert a.isdisjoint(c)

    assert a <= a
    assert a >= a
    assert not a < a
    assert not a > a
    assert not a <= b
    assert not a >= b
    assert not a >= c
    assert not a <= c
    assert c <= d
    assert c < d
    assert d >= c
    assert d > c


def test_schema_infer_pyarrow_table():
    table = pa.Table.from_arrays(
        [
            pa.array([1, 2, 3]),
            pa.array(["a", "b", "c"]),
            pa.array([True, False, True]),
        ],
        ["a", "b", "c"],
    )
    s = sch.infer(table)
    assert s == sch.Schema({"a": dt.int64, "b": dt.string, "c": dt.boolean})


def test_schema_from_to_pyarrow_schema():
    pyarrow_schema = pa.schema(
        [
            pa.field("a", pa.int64()),
            pa.field("b", pa.string()),
            pa.field("c", pa.bool_()),
        ]
    )
    ibis_schema = sch.schema(pyarrow_schema)
    restored_schema = ibis_schema.to_pyarrow()

    assert ibis_schema == sch.Schema({"a": dt.int64, "b": dt.string, "c": dt.boolean})
    assert restored_schema == pyarrow_schema


def test_schema_from_to_numpy_dtypes():
    numpy_dtypes = [
        ("a", np.dtype("int64")),
        ("b", np.dtype("str")),
        ("c", np.dtype("bool")),
    ]
    ibis_schema = sch.Schema.from_numpy(numpy_dtypes)
    assert ibis_schema == sch.Schema({"a": dt.int64, "b": dt.string, "c": dt.boolean})

    restored_dtypes = ibis_schema.to_numpy()
    expected_dtypes = [
        ("a", np.dtype("int64")),
        ("b", np.dtype("object")),
        ("c", np.dtype("bool")),
    ]
    assert restored_dtypes == expected_dtypes


@pytest.mark.parametrize(
    ("from_method", "to_method"),
    [
        pytest.param(
            "from_dask",
            "to_dask",
            marks=pytest.mark.skipif(not has_dask, reason="dask not installed"),
        ),
        pytest.param(
            "from_pandas",
            "to_pandas",
            marks=pytest.mark.skipif(not has_pandas, reason="pandas not installed"),
        ),
    ],
)
def test_schema_from_to_pandas_dask_dtypes(from_method, to_method):
    pandas_schema = pd.Series(
        [
            ("a", np.dtype("int64")),
            ("b", np.dtype("str")),
            ("c", pd.CategoricalDtype(["a", "b", "c"])),
            ("d", pd.DatetimeTZDtype(tz="US/Eastern", unit="ns")),
        ]
    )
    ibis_schema = getattr(sch.Schema, from_method)(pandas_schema)
    assert ibis_schema == sch.schema(pandas_schema)

    expected = sch.Schema(
        {
            "a": dt.int64,
            "b": dt.string,
            "c": dt.string,
            "d": dt.Timestamp(timezone="US/Eastern"),
        }
    )
    assert ibis_schema == expected

    restored_dtypes = getattr(ibis_schema, to_method)()
    expected_dtypes = [
        ("a", np.dtype("int64")),
        ("b", np.dtype("object")),
        ("c", np.dtype("object")),
        ("d", pd.DatetimeTZDtype(tz="US/Eastern", unit="ns")),
    ]
    assert restored_dtypes == expected_dtypes



from __future__ import annotations

import numpy as np
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis.backends.base.sqlglot.datatypes import DuckDBType


@pytest.mark.parametrize(
    ("typ", "expected"),
    [
        param(typ, expected, id=typ.lower())
        for typ, expected in [
            ("BIGINT", dt.int64),
            ("BOOLEAN", dt.boolean),
            ("BLOB", dt.binary),
            ("DATE", dt.date),
            ("DOUBLE", dt.float64),
            ("DECIMAL(10, 3)", dt.Decimal(10, 3)),
            ("INTEGER", dt.int32),
            ("INTERVAL", dt.Interval("us")),
            ("FLOAT", dt.float32),
            ("SMALLINT", dt.int16),
            ("TIME", dt.time),
            ("TIME WITH TIME ZONE", dt.time),
            ("TIMESTAMP", dt.Timestamp(scale=6)),
            ("TIMESTAMP WITH TIME ZONE", dt.Timestamp(scale=6, timezone="UTC")),
            ("TINYINT", dt.int8),
            ("UBIGINT", dt.uint64),
            ("UINTEGER", dt.uint32),
            ("USMALLINT", dt.uint16),
            ("UTINYINT", dt.uint8),
            ("UUID", dt.uuid),
            ("VARCHAR", dt.string),
            ("INTEGER[]", dt.Array(dt.int32)),
            ("MAP(VARCHAR, BIGINT)", dt.Map(dt.string, dt.int64)),
            (
                "STRUCT(a INTEGER, b VARCHAR, c MAP(VARCHAR, DOUBLE[])[])",
                dt.Struct(
                    dict(
                        a=dt.int32,
                        b=dt.string,
                        c=dt.Array(dt.Map(dt.string, dt.Array(dt.float64))),
                    )
                ),
            ),
            ("INTEGER[][]", dt.Array(dt.Array(dt.int32))),
            ("JSON", dt.json),
            ("HUGEINT", dt.Decimal(38, 0)),
            ("TIMESTAMP_S", dt.Timestamp(scale=0)),
            ("TIMESTAMP_MS", dt.Timestamp(scale=3)),
            ("TIMESTAMP_NS", dt.Timestamp(scale=9)),
        ]
    ],
)
def test_parser(typ, expected):
    ty = DuckDBType.from_string(typ)
    assert ty == expected


@pytest.mark.parametrize("uint_type", ["uint8", "uint16", "uint32", "uint64"])
def test_cast_uints(uint_type, snapshot):
    t = ibis.table(dict(a="int8"), name="t")
    snapshot.assert_match(
        str(ibis.to_sql(t.a.cast(uint_type), dialect="duckdb")), "out.sql"
    )


def test_null_dtype():
    con = ibis.connect("duckdb://:memory:")

    t = ibis.memtable({"a": [None, None]})
    assert t.schema() == ibis.schema(dict(a="null"))

    df = con.execute(t)
    assert df.a.tolist() == [None, None]


def test_parse_quoted_struct_field():
    assert DuckDBType.from_string('STRUCT("a" INTEGER, "a b c" INTEGER)') == dt.Struct(
        {"a": dt.int32, "a b c": dt.int32}
    )


def test_read_uint8_from_parquet(tmp_path):
    con = ibis.duckdb.connect()

    # There is an incorrect mapping in duckdb-engine from UInteger -> UInt8
    # In order to get something that reads as a UInt8, we cast to UInt32 (UInteger)
    t = ibis.memtable({"a": np.array([1, 2, 3, 4], dtype="uint32")})
    assert t.a.type() == dt.uint32

    parqpath = tmp_path / "uint.parquet"

    con.to_parquet(t, parqpath)

    # If this doesn't fail, then things are working
    t2 = con.read_parquet(parqpath)

    assert t2.schema() == t.schema()


@pytest.mark.parametrize("typ", ["float32", "float64"])
def test_cast_to_floating_point_type(con, snapshot, typ):
    expected = 1.0
    value = ibis.literal(str(expected))
    expr = value.cast(typ)

    result = con.execute(expr)
    assert result == expected

    sql = str(ibis.to_sql(expr, dialect="duckdb"))
    snapshot.assert_match(sql, "out.sql")



from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.base.sqlglot.compiler import NULL, SQLGlotCompiler
from ibis.backends.base.sqlglot.datatypes import ExasolType
from ibis.backends.base.sqlglot.dialects import Exasol
from ibis.backends.base.sqlglot.rewrites import (
    exclude_unsupported_window_frame_from_ops,
    exclude_unsupported_window_frame_from_rank,
    exclude_unsupported_window_frame_from_row_number,
    rewrite_empty_order_by_window,
    rewrite_sample_as_filter,
)


class ExasolCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Exasol
    type_mapper = ExasolType
    rewrites = (
        rewrite_sample_as_filter,
        exclude_unsupported_window_frame_from_ops,
        exclude_unsupported_window_frame_from_rank,
        exclude_unsupported_window_frame_from_row_number,
        rewrite_empty_order_by_window,
        *SQLGlotCompiler.rewrites,
    )

    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.AnalyticVectorizedUDF,
            ops.ApproxMedian,
            ops.Arbitrary,
            ops.ArgMax,
            ops.ArgMin,
            ops.ArrayCollect,
            ops.ArrayDistinct,
            ops.ArrayFilter,
            ops.ArrayFlatten,
            ops.ArrayIntersect,
            ops.ArrayMap,
            ops.ArraySort,
            ops.ArrayStringJoin,
            ops.ArrayUnion,
            ops.ArrayZip,
            ops.BitwiseNot,
            ops.Covariance,
            ops.CumeDist,
            ops.DateAdd,
            ops.DateDelta,
            ops.DateSub,
            ops.DateFromYMD,
            ops.DayOfWeekIndex,
            ops.DayOfWeekName,
            ops.ElementWiseVectorizedUDF,
            ops.ExtractDayOfYear,
            ops.ExtractEpochSeconds,
            ops.ExtractQuarter,
            ops.ExtractWeekOfYear,
            ops.First,
            ops.IntervalFromInteger,
            ops.IsInf,
            ops.IsNan,
            ops.Last,
            ops.Levenshtein,
            ops.Median,
            ops.MultiQuantile,
            ops.Quantile,
            ops.ReductionVectorizedUDF,
            ops.RegexExtract,
            ops.RegexReplace,
            ops.RegexSearch,
            ops.RegexSplit,
            ops.RowID,
            ops.StandardDev,
            ops.Strftime,
            ops.StringJoin,
            ops.StringSplit,
            ops.StringToTimestamp,
            ops.TimeDelta,
            ops.TimestampAdd,
            ops.TimestampBucket,
            ops.TimestampDelta,
            ops.TimestampDiff,
            ops.TimestampNow,
            ops.TimestampSub,
            ops.TimestampTruncate,
            ops.TypeOf,
            ops.Unnest,
            ops.Variance,
        )
    )

    SIMPLE_OPS = {
        ops.Log10: "log10",
        ops.Modulus: "mod",
        ops.All: "min",
        ops.Any: "max",
    }

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def _aggregate(self, funcname: str, *args, where):
        func = self.f[funcname]
        if where is not None:
            args = tuple(self.if_(where, arg, NULL) for arg in args)
        return func(*args)

    @staticmethod
    def _gen_valid_name(name: str) -> str:
        """Exasol does not allow dots in quoted column names."""
        return name.replace(".", "_")

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_timestamp():
            val = value.replace(tzinfo=None).isoformat(sep=" ", timespec="milliseconds")
            return self.cast(val, dtype)
        elif dtype.is_array() or dtype.is_struct() or dtype.is_map():
            raise com.UnsupportedBackendType(
                f"{type(dtype).__name__}s are not supported in Exasol"
            )
        elif dtype.is_uuid():
            return sge.convert(str(value))
        return super().visit_NonNullLiteral(op, value=value, dtype=dtype)

    def visit_Date(self, op, *, arg):
        return self.cast(arg, dt.date)

    def visit_StartsWith(self, op, *, arg, start):
        return self.f.left(arg, self.f.length(start)).eq(start)

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.length(end)).eq(end)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        return self.f.locate(substr, arg, (start if start is not None else 0) + 1)

    def visit_StringSQLILike(self, op, *, arg, pattern, escape):
        return self.f.upper(arg).like(self.f.upper(pattern))

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.locate(needle, haystack) > 0

    def visit_ExtractSecond(self, op, *, arg):
        return self.f.floor(self.cast(self.f.extract(self.v.second, arg), op.dtype))

    def visit_StringConcat(self, op, *, arg):
        any_args_null = (a.is_(NULL) for a in arg)
        return self.if_(sg.or_(*any_args_null), NULL, self.f.concat(*arg))

    def visit_CountDistinctStar(self, op, *, arg, where):
        raise com.UnsupportedOperationError(
            "COUNT(DISTINCT *) is not supported in Exasol"
        )

    def visit_DateTruncate(self, op, *, arg, unit):
        return super().visit_TimestampTruncate(op, arg=arg, unit=unit)



from __future__ import annotations

import operator

import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis.common.annotations import ValidationError


def test_type_metadata(lineitem):
    col = lineitem.l_extendedprice
    assert isinstance(col, ir.DecimalColumn)

    assert col.type() == dt.Decimal(12, 2)


def test_cast_scalar_to_decimal():
    val = ibis.literal("1.2345")

    casted = val.cast("decimal(15,5)")
    assert isinstance(casted, ir.DecimalScalar)
    assert casted.type() == dt.Decimal(15, 5)


def test_decimal_sum_type(lineitem):
    col = lineitem.l_extendedprice
    result = col.sum()
    assert isinstance(result, ir.DecimalScalar)
    assert result.type() == dt.Decimal(38, 2)


def test_promote_decimal_type_mul(lineitem):
    col_1 = lineitem.l_extendedprice
    col_2 = lineitem.l_discount
    result = col_1 * col_2
    assert result.type().precision == 12
    assert result.type().scale == 2


def test_promote_decimal_type_add(lineitem):
    col_1 = lineitem.l_extendedprice
    col_2 = lineitem.l_discount
    result = col_1 + col_2
    assert result.type().precision == 12
    assert result.type().scale == 2


def test_promote_decimal_type_mod(lineitem):
    col_1 = lineitem.l_extendedprice
    col_2 = lineitem.l_discount
    result = col_1 % col_2
    assert result.type().precision == 12
    assert result.type().scale == 2


def test_promote_decimal_type_max():
    t = ibis.table([("a", "decimal(31, 3)"), ("b", "decimal(31, 3)")], "t")
    result = t.a * t.b
    assert result.type().precision == 31
    assert result.type().scale == 3


@pytest.mark.parametrize(
    "precision, scale, expected",
    [
        (None, None, (None, None)),
        (38, 2, (38, 2)),
        (16, 2, (38, 2)),
        (39, 3, (39, 3)),
    ],
)
def test_decimal_sum_type_precision(precision, scale, expected):
    t = ibis.table([("l_extendedprice", dt.Decimal(precision, scale))], name="t")
    col = t.l_extendedprice
    result = col.sum()
    assert isinstance(result, ir.DecimalScalar)
    assert result.type() == dt.Decimal(*expected)


@pytest.mark.parametrize("func", ["mean", "max", "min"])
def test_decimal_aggregate_function_type(lineitem, func):
    col = lineitem.l_extendedprice
    method = operator.methodcaller(func)
    result = method(col)
    assert isinstance(result, ir.DecimalScalar)
    assert result.type() == col.type()


def test_ifelse(lineitem):
    table = lineitem

    q = table.l_quantity
    expr = ibis.ifelse(table.l_discount > 0, q * table.l_discount, ibis.null())

    assert isinstance(expr, ir.DecimalColumn)

    expr = ibis.ifelse(table.l_discount > 0, (q * table.l_discount).sum(), ibis.null())
    assert isinstance(expr, ir.DecimalColumn)

    expr = ibis.ifelse(
        table.l_discount.sum() > 0, (q * table.l_discount).sum(), ibis.null()
    )
    assert isinstance(expr, ir.DecimalScalar)


def test_fillna(lineitem):
    expr = lineitem.l_extendedprice.fillna(0)
    assert isinstance(expr, ir.DecimalColumn)

    expr = lineitem.l_extendedprice.fillna(lineitem.l_quantity)
    assert isinstance(expr, ir.DecimalColumn)


@pytest.mark.parametrize(
    ("precision", "scale"),
    [
        (-1, 3),  # negative precision
        (0, 1),  # zero precision
        (12, 38),  # precision less than scale
        (33, -1),  # negative scale
    ],
)
def test_invalid_precision_scale_combo(precision, scale):
    with pytest.raises(ValueError):
        dt.Decimal(precision, scale)


@pytest.mark.parametrize(
    ("precision", "scale"),
    [(38.1, 3), (38, 3.1)],  # non integral precision  # non integral scale
)
def test_invalid_precision_scale_type(precision, scale):
    with pytest.raises(ValidationError):
        dt.Decimal(precision, scale)


def test_decimal_str(lineitem):
    col = lineitem.l_extendedprice
    t = col.type()
    assert str(t) == f"decimal({t.precision:d}, {t.scale:d})"


def test_decimal_repr(lineitem):
    col = lineitem.l_extendedprice
    t = col.type()
    expected = f"Decimal(precision={t.precision:d}, scale={t.scale:d}, nullable=True)"
    assert repr(t) == expected



from __future__ import annotations

try:
    from duckdb import ConversionException as DuckDBConversionException
    from duckdb import InvalidInputException as DuckDBInvalidInputException
    from duckdb import NotImplementedException as DuckDBNotImplementedException
    from duckdb import ParserException as DuckDBParserException
except ImportError:
    DuckDBConversionException = (
        DuckDBInvalidInputException
    ) = DuckDBParserException = DuckDBNotImplementedException = None

try:
    from clickhouse_connect.driver.exceptions import (
        DatabaseError as ClickHouseDatabaseError,
    )
    from clickhouse_connect.driver.exceptions import (
        InternalError as ClickHouseInternalError,
    )
    from clickhouse_connect.driver.exceptions import (
        OperationalError as ClickHouseOperationalError,
    )
except ImportError:
    ClickHouseDatabaseError = (
        ClickHouseInternalError
    ) = ClickHouseOperationalError = None


try:
    from pyexasol.exceptions import ExaQueryError
except ImportError:
    ExaQueryError = None

try:
    from pyspark.sql.utils import AnalysisException as PySparkAnalysisException
    from pyspark.sql.utils import (
        IllegalArgumentException as PySparkIllegalArgumentException,
    )
    from pyspark.sql.utils import ParseException as PySparkParseException
    from pyspark.sql.utils import PythonException as PySparkPythonException
except ImportError:
    PySparkAnalysisException = (
        PySparkIllegalArgumentException
    ) = PySparkParseException = PySparkPythonException = None

try:
    # PySpark 3.5.0
    from pyspark.errors.exceptions.captured import (
        ArithmeticException as PySparkArithmeticException,
    )
except ImportError:
    PySparkArithmeticException = None

try:
    from google.api_core.exceptions import BadRequest as GoogleBadRequest
except ImportError:
    GoogleBadRequest = None

try:
    from polars import ComputeError as PolarsComputeError
    from polars import PanicException as PolarsPanicException
    from polars.exceptions import InvalidOperationError as PolarsInvalidOperationError
except ImportError:
    PolarsComputeError = PolarsPanicException = PolarsInvalidOperationError = None

try:
    from pyarrow import ArrowInvalid, ArrowNotImplementedError
except ImportError:
    ArrowInvalid = ArrowNotImplementedError = None

try:
    from impala.error import HiveServer2Error as ImpalaHiveServer2Error
    from impala.error import OperationalError as ImpalaOperationalError
except ImportError:
    ImpalaHiveServer2Error = ImpalaOperationalError = None

try:
    from py4j.protocol import Py4JError, Py4JJavaError
except ImportError:
    Py4JJavaError = Py4JError = None

try:
    from deltalake import PyDeltaTableError
except ImportError:
    PyDeltaTableError = None

try:
    from snowflake.connector.errors import ProgrammingError as SnowflakeProgrammingError
except ImportError:
    SnowflakeProgrammingError = None

try:
    from trino.exceptions import TrinoUserError
except ImportError:
    TrinoUserError = None

try:
    from psycopg2.errors import DivisionByZero as PsycoPg2DivisionByZero
    from psycopg2.errors import IndeterminateDatatype as PsycoPg2IndeterminateDatatype
    from psycopg2.errors import InternalError_ as PsycoPg2InternalError
    from psycopg2.errors import (
        InvalidTextRepresentation as PsycoPg2InvalidTextRepresentation,
    )
    from psycopg2.errors import OperationalError as PsycoPg2OperationalError
    from psycopg2.errors import ProgrammingError as PsycoPg2ProgrammingError
    from psycopg2.errors import SyntaxError as PsycoPg2SyntaxError
    from psycopg2.errors import UndefinedObject as PsycoPg2UndefinedObject
except ImportError:
    PsycoPg2SyntaxError = (
        PsycoPg2IndeterminateDatatype
    ) = (
        PsycoPg2InvalidTextRepresentation
    ) = (
        PsycoPg2DivisionByZero
    ) = (
        PsycoPg2InternalError
    ) = (
        PsycoPg2ProgrammingError
    ) = PsycoPg2OperationalError = PsycoPg2UndefinedObject = None

try:
    from pymysql.err import NotSupportedError as MySQLNotSupportedError
    from pymysql.err import OperationalError as MySQLOperationalError
    from pymysql.err import ProgrammingError as MySQLProgrammingError
except ImportError:
    MySQLNotSupportedError = MySQLProgrammingError = MySQLOperationalError = None

try:
    from pydruid.db.exceptions import ProgrammingError as PyDruidProgrammingError
except ImportError:
    PyDruidProgrammingError = None

try:
    from oracledb.exceptions import DatabaseError as OracleDatabaseError
except ImportError:
    OracleDatabaseError = None

try:
    from pyodbc import DataError as PyODBCDataError
    from pyodbc import ProgrammingError as PyODBCProgrammingError
except ImportError:
    PyODBCProgrammingError = PyODBCDataError = None



import ibis


alltypes = ibis.table(
    name="alltypes",
    schema={
        "a": "int8",
        "b": "int16",
        "c": "int32",
        "d": "int64",
        "e": "float32",
        "f": "float64",
        "g": "string",
        "h": "boolean",
        "i": "timestamp",
        "j": "date",
        "k": "time",
    },
)
lit = ibis.literal("foo")
lit1 = ibis.literal("baz")
lit2 = ibis.literal("bar")

result = alltypes.select(
    alltypes.g.case()
    .when(lit, lit2)
    .when(lit1, ibis.literal("qux"))
    .else_(ibis.literal("default"))
    .end()
    .name("col1"),
    ibis.case()
    .when(alltypes.g == lit, lit2)
    .when(alltypes.g == lit1, alltypes.g)
    .else_(ibis.literal(None).cast("string"))
    .end()
    .name("col2"),
    alltypes.a,
    alltypes.b,
    alltypes.c,
    alltypes.d,
    alltypes.e,
    alltypes.f,
    alltypes.g,
    alltypes.h,
    alltypes.i,
    alltypes.j,
    alltypes.k,
)



import ibis


employee = ibis.table(
    name="employee",
    schema={"first_name": "string", "last_name": "string", "id": "int64"},
)
f = employee.filter(employee.id < 5)
s = f.order_by(f.id.desc())

result = s.select(s.first_name, s.last_name, s.id, s.first_name.name("first"))



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from ibis.formats.pandas import PandasData
from ibis.formats.pyarrow import PYARROW_JSON_TYPE, PyArrowData

if TYPE_CHECKING:
    import pyarrow as pa

    import ibis.expr.datatypes as dt
    from ibis.expr.schema import Schema


class SnowflakePandasData(PandasData):
    @classmethod
    def convert_Timestamp_element(cls, dtype):
        return datetime.datetime.fromisoformat

    @classmethod
    def convert_Date_element(cls, dtype):
        return datetime.date.fromisoformat

    @classmethod
    def convert_Time_element(cls, dtype):
        return datetime.time.fromisoformat

    @classmethod
    def convert_JSON(cls, s, dtype, pandas_type):
        converter = cls.convert_JSON_element(dtype)
        return s.map(converter, na_action="ignore").astype("object")

    @classmethod
    def convert_Array(cls, s, dtype, pandas_type):
        raw_json_objects = cls.convert_JSON(s, dtype, pandas_type)
        return super().convert_Array(raw_json_objects, dtype, pandas_type)

    @classmethod
    def convert_Map(cls, s, dtype, pandas_type):
        raw_json_objects = cls.convert_JSON(s, dtype, pandas_type)
        return super().convert_Map(raw_json_objects, dtype, pandas_type)

    @classmethod
    def convert_Struct(cls, s, dtype, pandas_type):
        raw_json_objects = cls.convert_JSON(s, dtype, pandas_type)
        return super().convert_Struct(raw_json_objects, dtype, pandas_type)


class SnowflakePyArrowData(PyArrowData):
    @classmethod
    def convert_table(cls, table: pa.Table, schema: Schema) -> pa.Table:
        import pyarrow as pa

        columns = [cls.convert_column(table[name], typ) for name, typ in schema.items()]
        return pa.Table.from_arrays(columns, names=schema.names)

    @classmethod
    def convert_column(cls, column: pa.Array, dtype: dt.DataType) -> pa.Array:
        if dtype.is_json() or dtype.is_array() or dtype.is_map() or dtype.is_struct():
            import pyarrow as pa

            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()

            return pa.ExtensionArray.from_storage(PYARROW_JSON_TYPE, column)
        return super().convert_column(column, dtype)



"""Convenient column selectors.

::: {.callout-tip}
## Check out the [blog post on selectors](../posts/selectors) for examples!
:::

## Rationale

Column selectors are convenience functions for selecting columns that share some property.

## Discussion

For example, a common task is to be able to select all numeric columns for a
subsequent computation.

Without selectors this becomes quite verbose and tedious to write:

>>> import ibis
>>> t = ibis.table(dict(a="int", b="string", c="array<int>", abcd="float"))
>>> expr = t.select([t[c] for c in t.columns if t[c].type().is_numeric()])
>>> expr.columns
['a', 'abcd']

Compare that to the [`numeric`](#ibis.selectors.numeric) selector:

>>> import ibis.selectors as s
>>> expr = t.select(s.numeric())
>>> expr.columns
['a', 'abcd']

When there are multiple properties to check it gets worse:

>>> expr = t.select(
...     [
...         t[c]
...         for c in t.columns
...         if t[c].type().is_numeric() or t[c].type().is_string()
...         if ("a" in c or "b" in c or "cd" in c)
...     ]
... )
>>> expr.columns
['a', 'b', 'abcd']

Using a composition of selectors this is much less tiresome:

>>> expr = t.select((s.numeric() | s.of_type("string")) & s.contains(("a", "b", "cd")))
>>> expr.columns
['a', 'b', 'abcd']
"""

from __future__ import annotations

import abc
import functools
import inspect
import operator
import re
from collections.abc import Iterable, Mapping, Sequence
from typing import Callable, Optional, Union

from public import public

import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis import util
from ibis.common.collections import frozendict  # noqa: TCH001
from ibis.common.deferred import Deferred, Resolver
from ibis.common.grounds import Concrete, Singleton


class Selector(Concrete):
    """A column selector."""

    @abc.abstractmethod
    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        """Expand `table` into value expressions that match the selector.

        Parameters
        ----------
        table
            An ibis table expression

        Returns
        -------
        Sequence[Value]
            A sequence of value expressions that match the selector

        """

    def positions(self, table: ir.Table) -> Sequence[int]:
        """Expand `table` into column indices that match the selector.

        Parameters
        ----------
        table
            An ibis table expression

        Returns
        -------
        Sequence[int]
            A sequence of column indices where the selector matches

        """
        raise NotImplementedError(
            f"`positions` doesn't make sense for {self.__class__.__name__} selector"
        )


class Predicate(Selector):
    predicate: Callable[[ir.Value], bool]

    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        """Evaluate `self.predicate` on every column of `table`.

        Parameters
        ----------
        table
            An ibis table expression

        """
        return [col for column in table.columns if self.predicate(col := table[column])]

    def positions(self, table: ir.Table) -> Sequence[int]:
        return [
            i for i, column in enumerate(table.columns) if self.predicate(table[column])
        ]

    def __and__(self, other: Selector) -> Predicate:
        """Compute the conjunction of two `Selector`s.

        Parameters
        ----------
        other
            Another selector

        """
        return self.__class__(lambda col: self.predicate(col) and other.predicate(col))

    def __or__(self, other: Selector) -> Predicate:
        """Compute the disjunction of two `Selector`s.

        Parameters
        ----------
        other
            Another selector

        """
        return self.__class__(lambda col: self.predicate(col) or other.predicate(col))

    def __invert__(self) -> Predicate:
        """Compute the logical negation of two `Selector`s."""
        return self.__class__(lambda col: not self.predicate(col))


@public
def where(predicate: Callable[[ir.Value], bool]) -> Predicate:
    """Select columns that satisfy `predicate`.

    Use this selector when one of the other selectors does not meet your needs.

    Parameters
    ----------
    predicate
        A callable that accepts an ibis value expression and returns a `bool`

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(a="float32"), name="t")
    >>> expr = t.select(s.where(lambda col: col.get_name() == "a"))
    >>> expr.columns
    ['a']

    """
    return Predicate(predicate=predicate)


@public
def numeric() -> Predicate:
    """Return numeric columns.

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(a="int", b="string", c="array<string>"), name="t")
    >>> t.columns
    ['a', 'b', 'c']
    >>> expr = t.select(s.numeric())  # `a` has integer type, so it's numeric
    >>> expr.columns
    ['a']

    See Also
    --------
    [`of_type`](#ibis.selectors.of_type)

    """
    return of_type(dt.Numeric)


@public
def of_type(dtype: dt.DataType | str | type[dt.DataType]) -> Predicate:
    """Select columns of type `dtype`.

    Parameters
    ----------
    dtype
        `DataType` instance, `str` or `DataType` class

    Examples
    --------
    Select according to a specific `DataType` instance

    >>> import ibis
    >>> import ibis.expr.datatypes as dt
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(name="string", siblings="array<string>", parents="array<int64>"))
    >>> expr = t.select(s.of_type(dt.Array(dt.string)))
    >>> expr.columns
    ['siblings']

    Strings are also accepted

    >>> expr = t.select(s.of_type("array<string>"))
    >>> expr.columns
    ['siblings']

    Abstract/unparametrized types may also be specified by their string name
    (e.g. "integer" for any integer type), or by passing in a `DataType` class
    instead. The following options are equivalent.

    >>> expr1 = t.select(s.of_type("array"))
    >>> expr2 = t.select(s.of_type(dt.Array))
    >>> expr1.equals(expr2)
    True
    >>> expr2.columns
    ['siblings', 'parents']

    See Also
    --------
    [`numeric`](#ibis.selectors.numeric)

    """
    if isinstance(dtype, str):
        # A mapping of abstract or parametric types, to allow selecting all
        # subclasses/parametrizations of these types, rather than only a
        # specific instance.
        abstract = {
            "array": dt.Array,
            "decimal": dt.Decimal,
            "floating": dt.Floating,
            "geospatial": dt.GeoSpatial,
            "integer": dt.Integer,
            "map": dt.Map,
            "numeric": dt.Numeric,
            "struct": dt.Struct,
            "temporal": dt.Temporal,
        }
        if cls := abstract.get(dtype.lower()):
            predicate = lambda col: isinstance(col.type(), cls)
        else:
            dtype = dt.dtype(dtype)
            predicate = lambda col: col.type() == dtype
    elif inspect.isclass(dtype):
        predicate = lambda col: isinstance(col.type(), dtype)
    else:
        dtype = dt.dtype(dtype)
        predicate = lambda col: col.type() == dtype
    return where(predicate)


@public
def startswith(prefixes: str | tuple[str, ...]) -> Predicate:
    """Select columns whose name starts with one of `prefixes`.

    Parameters
    ----------
    prefixes
        Prefixes to compare column names against

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(apples="int", oranges="float", bananas="bool"), name="t")
    >>> expr = t.select(s.startswith(("a", "b")))
    >>> expr.columns
    ['apples', 'bananas']

    See Also
    --------
    [`endswith`](#ibis.selectors.endswith)

    """
    return where(lambda col: col.get_name().startswith(prefixes))


@public
def endswith(suffixes: str | tuple[str, ...]) -> Predicate:
    """Select columns whose name ends with one of `suffixes`.

    Parameters
    ----------
    suffixes
        Suffixes to compare column names against

    See Also
    --------
    [`startswith`](#ibis.selectors.startswith)

    """
    return where(lambda col: col.get_name().endswith(suffixes))


@public
def contains(
    needles: str | tuple[str, ...], how: Callable[[Iterable[bool]], bool] = any
) -> Predicate:
    """Return columns whose name contains `needles`.

    Parameters
    ----------
    needles
        One or more strings to search for in column names
    how
        A boolean reduction to allow the configuration of how `needles` are summarized.

    Examples
    --------
    Select columns that contain either `"a"` or `"b"`

    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(
    ...     dict(a="int64", b="string", c="float", d="array<int16>", ab="struct<x: int>")
    ... )
    >>> expr = t.select(s.contains(("a", "b")))
    >>> expr.columns
    ['a', 'b', 'ab']

    Select columns that contain all of `"a"` and `"b"`, that is, both `"a"` and
    `"b"` must be in each column's name to match.

    >>> expr = t.select(s.contains(("a", "b"), how=all))
    >>> expr.columns
    ['ab']

    See Also
    --------
    [`matches`](#ibis.selectors.matches)

    """

    def predicate(col: ir.Value) -> bool:
        name = col.get_name()
        return how(needle in name for needle in util.promote_list(needles))

    return where(predicate)


@public
def matches(regex: str | re.Pattern) -> Selector:
    """Return columns whose name matches the regular expression `regex`.

    Parameters
    ----------
    regex
        A string or `re.Pattern` object

    Examples
    --------
    >>> import ibis
    >>> import ibis.selectors as s
    >>> t = ibis.table(dict(ab="string", abd="int", be="array<string>"))
    >>> expr = t.select(s.matches(r"ab+"))
    >>> expr.columns
    ['ab', 'abd']

    See Also
    --------
    [`contains`](#ibis.selectors.contains)

    """
    pattern = re.compile(regex)
    return where(lambda col: pattern.search(col.get_name()) is not None)


@public
def any_of(*predicates: str | Predicate) -> Predicate:
    """Include columns satisfying any of `predicates`."""
    return functools.reduce(operator.or_, map(_to_selector, predicates))


@public
def all_of(*predicates: str | Predicate) -> Predicate:
    """Include columns satisfying all of `predicates`."""
    return functools.reduce(operator.and_, map(_to_selector, predicates))


@public
def c(*names: str | ir.Column) -> Predicate:
    """Select specific column names."""
    names = frozenset(col if isinstance(col, str) else col.get_name() for col in names)

    def func(col: ir.Value) -> bool:
        schema = col.op().rel.schema
        if extra_cols := (names - schema.keys()):
            raise exc.IbisInputError(
                f"Columns {extra_cols} are not present in {schema.names}"
            )
        return col.get_name() in names

    return where(func)


class Across(Selector):
    selector: Selector
    funcs: Union[
        Resolver,
        Callable[[ir.Value], ir.Value],
        frozendict[Optional[str], Union[Resolver, Callable[[ir.Value], ir.Value]]],
    ]
    names: Union[str, Callable[[str, Optional[str]], str]]

    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        expanded = []

        names = self.names
        cols = self.selector.expand(table)
        for func_name, func in self.funcs.items():
            for orig_col in cols:
                if isinstance(func, Resolver):
                    col = func.resolve({"_": orig_col})
                else:
                    col = func(orig_col)

                if callable(names):
                    name = names(orig_col.get_name(), func_name)
                else:
                    name = names.format(col=orig_col.get_name(), fn=func_name)

                expanded.append(col.name(name))

        return expanded


@public
def across(
    selector: Selector | Iterable[str] | str,
    func: Deferred
    | Callable[[ir.Value], ir.Value]
    | Mapping[str | None, Deferred | Callable[[ir.Value], ir.Value]],
    names: str | Callable[[str, str | None], str] | None = None,
) -> Across:
    """Apply data transformations across multiple columns.

    Parameters
    ----------
    selector
        An expression that selects columns on which the transformation function
        will be applied, an iterable of `str` column names or a single `str`
        column name.
    func
        A function (or dictionary of functions) to use to transform the data.
    names
        A lambda function or a format string to name the columns created by the
        transformation function.

    Returns
    -------
    Across
        An `Across` selector object

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> from ibis import _, selectors as s
    >>> t = ibis.examples.penguins.fetch()
    >>> t.select(s.startswith("bill")).mutate(
    ...     s.across(s.numeric(), dict(centered=_ - _.mean()), names="{fn}_{col}")
    ... )
    ┏━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ bill_length_mm ┃ bill_depth_mm ┃ centered_bill_length_mm ┃ … ┃
    ┡━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ float64        │ float64       │ float64                 │ … │
    ├────────────────┼───────────────┼─────────────────────────┼───┤
    │           39.1 │          18.7 │                -4.82193 │ … │
    │           39.5 │          17.4 │                -4.42193 │ … │
    │           40.3 │          18.0 │                -3.62193 │ … │
    │           NULL │          NULL │                    NULL │ … │
    │           36.7 │          19.3 │                -7.22193 │ … │
    │           39.3 │          20.6 │                -4.62193 │ … │
    │           38.9 │          17.8 │                -5.02193 │ … │
    │           39.2 │          19.6 │                -4.72193 │ … │
    │           34.1 │          18.1 │                -9.82193 │ … │
    │           42.0 │          20.2 │                -1.92193 │ … │
    │              … │             … │                       … │ … │
    └────────────────┴───────────────┴─────────────────────────┴───┘

    """
    if names is None:
        names = lambda col, fn: "_".join(filter(None, (col, fn)))
    funcs = dict(func if isinstance(func, Mapping) else {None: func})
    if not isinstance(selector, Selector):
        selector = c(*util.promote_list(selector))
    return Across(selector=selector, funcs=funcs, names=names)


class IfAnyAll(Selector):
    selector: Selector
    predicate: Union[Resolver, Callable[[ir.Value], ir.BooleanValue]]
    summarizer: Callable[[ir.BooleanValue, ir.BooleanValue], ir.BooleanValue]

    def expand(self, table: ir.Table) -> Sequence[ir.Value]:
        func = self.predicate
        if isinstance(func, Resolver):
            elems = (func.resolve({"_": col}) for col in self.selector.expand(table))
        else:
            elems = (func(col) for col in self.selector.expand(table))

        return [functools.reduce(self.summarizer, elems)]


@public
def if_any(selector: Selector, predicate: Deferred | Callable) -> IfAnyAll:
    """Return the **disjunction** of `predicate` applied on all `selector` columns.

    Parameters
    ----------
    selector
        A column selector
    predicate
        A callable or deferred object defining a predicate to apply to each
        column from `selector`.

    Examples
    --------
    >>> import ibis
    >>> from ibis import selectors as s, _
    >>> ibis.options.interactive = True
    >>> penguins = ibis.examples.penguins.fetch()
    >>> cols = s.across(s.endswith("_mm"), (_ - _.mean()) / _.std())
    >>> expr = penguins.mutate(cols).filter(s.if_any(s.endswith("_mm"), _.abs() > 2))
    >>> expr_by_hand = penguins.mutate(cols).filter(
    ...     (_.bill_length_mm.abs() > 2)
    ...     | (_.bill_depth_mm.abs() > 2)
    ...     | (_.flipper_length_mm.abs() > 2)
    ... )
    >>> expr.equals(expr_by_hand)
    True
    >>> expr
    ┏━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ species ┃ island ┃ bill_length_mm ┃ bill_depth_mm ┃ flipper_length_mm ┃ … ┃
    ┡━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ string  │ string │ float64        │ float64       │ float64           │ … │
    ├─────────┼────────┼────────────────┼───────────────┼───────────────────┼───┤
    │ Adelie  │ Biscoe │      -1.103002 │      0.733662 │         -2.056307 │ … │
    │ Gentoo  │ Biscoe │       1.113285 │     -0.431017 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       2.871660 │     -0.076550 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       1.900890 │     -0.734846 │          2.139483 │ … │
    │ Gentoo  │ Biscoe │       1.076652 │     -0.177826 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       0.856855 │     -0.582932 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       1.497929 │     -0.076550 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       1.388031 │     -0.431017 │          2.068368 │ … │
    │ Gentoo  │ Biscoe │       2.047422 │     -0.582932 │          2.068368 │ … │
    │ Adelie  │ Dream  │      -2.165354 │     -0.836123 │         -0.918466 │ … │
    │ …       │ …      │              … │             … │                 … │ … │
    └─────────┴────────┴────────────────┴───────────────┴───────────────────┴───┘

    """
    return IfAnyAll(selector=selector, predicate=predicate, summarizer=operator.or_)


@public
def if_all(selector: Selector, predicate: Deferred | Callable) -> IfAnyAll:
    """Return the **conjunction** of `predicate` applied on all `selector` columns.

    Parameters
    ----------
    selector
        A column selector
    predicate
        A callable or deferred object defining a predicate to apply to each
        column from `selector`.

    Examples
    --------
    >>> import ibis
    >>> from ibis import selectors as s, _
    >>> ibis.options.interactive = True
    >>> penguins = ibis.examples.penguins.fetch()
    >>> cols = s.across(s.endswith("_mm"), (_ - _.mean()) / _.std())
    >>> expr = penguins.mutate(cols).filter(s.if_all(s.endswith("_mm"), _.abs() > 1))
    >>> expr_by_hand = penguins.mutate(cols).filter(
    ...     (_.bill_length_mm.abs() > 1)
    ...     & (_.bill_depth_mm.abs() > 1)
    ...     & (_.flipper_length_mm.abs() > 1)
    ... )
    >>> expr.equals(expr_by_hand)
    True
    >>> expr
    ┏━━━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ species ┃ island    ┃ bill_length_mm ┃ bill_depth_mm ┃ flipper_length_mm ┃ … ┃
    ┡━━━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ string  │ string    │ float64        │ float64       │ float64           │ … │
    ├─────────┼───────────┼────────────────┼───────────────┼───────────────────┼───┤
    │ Adelie  │ Dream     │      -1.157951 │      1.088129 │         -1.416272 │ … │
    │ Adelie  │ Torgersen │      -1.231217 │      1.138768 │         -1.202926 │ … │
    │ Gentoo  │ Biscoe    │       1.149917 │     -1.443781 │          1.214987 │ … │
    │ Gentoo  │ Biscoe    │       1.040019 │     -1.089314 │          1.072757 │ … │
    │ Gentoo  │ Biscoe    │       1.131601 │     -1.089314 │          1.712792 │ … │
    │ Gentoo  │ Biscoe    │       1.241499 │     -1.089314 │          1.570562 │ … │
    │ Gentoo  │ Biscoe    │       1.351398 │     -1.494420 │          1.214987 │ … │
    └─────────┴───────────┴────────────────┴───────────────┴───────────────────┴───┘

    """
    return IfAnyAll(selector=selector, predicate=predicate, summarizer=operator.and_)


class Sliceable(Singleton):
    def __getitem__(self, key: str | int | slice | Iterable[int | str]) -> Predicate:
        import ibis.expr.analysis as an

        def pred(col: ir.Value) -> bool:
            table = an.find_first_base_table(col.op())
            schema = table.schema
            idxs = schema._name_locs
            num_names = len(schema)
            colname = col.get_name()
            colidx = idxs[colname]

            if isinstance(key, str):
                return key == colname
            elif isinstance(key, int):
                return key % num_names == colidx
            elif util.is_iterable(key):
                return any(
                    (isinstance(el, int) and el % num_names == colidx)
                    or (isinstance(el, str) and el == colname)
                    for el in key
                )
            else:
                start = key.start or 0
                stop = key.stop or num_names
                step = key.step or 1

                if isinstance(start, str):
                    start = idxs[start]

                if isinstance(stop, str):
                    stop = idxs[stop] + 1

                return colidx in range(start, stop, step)

        return where(pred)


r = Sliceable()
"""Ranges of columns."""


@public
def first() -> Predicate:
    """Return the first column of a table."""
    return r[0]


@public
def last() -> Predicate:
    """Return the last column of a table."""
    return r[-1]


@public
def all() -> Predicate:
    """Return every column from a table."""
    return r[:]


def _to_selector(
    obj: str | Selector | ir.Column | Sequence[str | Selector | ir.Column],
) -> Selector:
    """Convert an object to a `Selector`."""
    if isinstance(obj, Selector):
        return obj
    elif isinstance(obj, ir.Column):
        return c(obj.get_name())
    elif isinstance(obj, str):
        return c(obj)
    else:
        return any_of(*obj)



from __future__ import annotations

from typing import TYPE_CHECKING

from public import public

import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis.expr.types.relations import bind

if TYPE_CHECKING:
    from ibis.expr.types import Table


@public
class WindowedTable:
    """An intermediate table expression to hold windowing information."""

    def __init__(self, table: ir.Table, time_col: ir.Value):
        self.table = table
        self.time_col = next(bind(table, time_col))

        if self.time_col is None:
            raise com.IbisInputError(
                "Window aggregations require `time_col` as an argument"
            )

    def tumble(
        self,
        window_size: ir.IntervalScalar,
        offset: ir.IntervalScalar | None = None,
    ) -> Table:
        """Compute a tumble table valued function.

        Tumbling windows have a fixed size and do not overlap. The size of the windows is
        determined by `window_size`, optionally shifted by a duration specified by `offset`.

        Parameters
        ----------
        window_size
            Width of the tumbling windows.
        offset
            An optional parameter to specify the offset which window start should be shifted by.

        Returns
        -------
        Table
            Table expression after applying tumbling table-valued function.
        """
        time_col = next(bind(self.table, self.time_col))
        return ops.TumbleWindowingTVF(
            table=self.table,
            time_col=time_col,
            window_size=window_size,
            offset=offset,
        ).to_expr()

    def hop(
        self,
        window_size: ir.IntervalScalar,
        window_slide: ir.IntervalScalar,
        offset: ir.IntervalScalar | None = None,
    ):
        """Compute a hop table valued function.

        Hopping windows have a fixed size and can be overlapping if the slide is smaller than the
        window size (in which case elements can be assigned to multiple windows). Hopping windows
        are also known as sliding windows. The size of the windows is determined by `window_size`,
        how frequently a hopping window is started is determined by `window_slide`, and windows can
        be optionally shifted by a duration specified by `offset`.

        For example, you could have windows of size 10 minutes that slides by 5 minutes. With this,
        you get every 5 minutes a window that contains the events that arrived during the last 10 minutes.

        Parameters
        ----------
        window_size
            Width of the hopping windows.
        window_slide
            The duration between the start of sequential hopping windows.
        offset
            An optional parameter to specify the offset which window start should be shifted by.

        Returns
        -------
        Table
            Table expression after applying hopping table-valued function.
        """
        time_col = next(bind(self.table, self.time_col))
        return ops.HopWindowingTVF(
            table=self.table,
            time_col=time_col,
            window_size=window_size,
            window_slide=window_slide,
            offset=offset,
        ).to_expr()

    def cumulate(
        self,
        window_size: ir.IntervalScalar,
        window_step: ir.IntervalScalar,
        offset: ir.IntervalScalar | None = None,
    ):
        """Compute a cumulate table valued function.

        Cumulate windows don't have a fixed size and do overlap. Cumulate windows assign elements to windows
        that cover rows within an initial interval of step size and expand to one more step size (keep window
        start fixed) every step until the max window size.

        For example, you could have a cumulating window for 1 hour step and 1 day max size, and you will get
        windows: [00:00, 01:00), [00:00, 02:00), [00:00, 03:00), …, [00:00, 24:00) for every day.

        Parameters
        ----------
        window_size
            Max width of the cumulating windows.
        window_step
            A duration specifying the increased window size between the end of sequential cumulating windows.
        offset
            An optional parameter to specify the offset which window start should be shifted by.

        Returns
        -------
        Table
            Table expression after applying cumulate table-valued function.
        """
        time_col = next(bind(self.table, self.time_col))
        return ops.CumulateWindowingTVF(
            table=self.table,
            time_col=time_col,
            window_size=window_size,
            window_step=window_step,
            offset=offset,
        ).to_expr()



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Any

import pandas as pd
import pyarrow as pa
import pyarrow_hotfix  # noqa: F401

import ibis.common.exceptions as com
import ibis.config
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base import BaseBackend, NoUrl
from ibis.formats.pandas import PandasData, PandasSchema
from ibis.formats.pyarrow import PyArrowData

if TYPE_CHECKING:
    import pathlib
    from collections.abc import Mapping, MutableMapping


class BasePandasBackend(BaseBackend, NoUrl):
    """Base class for backends based on pandas."""

    name = "pandas"
    dialect = None
    backend_table_type = pd.DataFrame

    class Options(ibis.config.Config):
        enable_trace: bool = False

    def do_connect(
        self,
        dictionary: MutableMapping[str, pd.DataFrame] | None = None,
    ) -> None:
        """Construct a client from a dictionary of pandas DataFrames.

        Parameters
        ----------
        dictionary
            An optional mapping of string table names to pandas DataFrames.

        Examples
        --------
        >>> import ibis
        >>> ibis.pandas.connect({"t": pd.DataFrame({"a": [1, 2, 3]})})
        <ibis.backends.pandas.Backend at 0x...>

        """
        self.dictionary = dictionary or {}
        self.schemas: MutableMapping[str, sch.Schema] = {}

    def disconnect(self) -> None:
        pass

    def from_dataframe(
        self,
        df: pd.DataFrame,
        name: str = "df",
        client: BasePandasBackend | None = None,
    ) -> ir.Table:
        """Construct an ibis table from a pandas DataFrame.

        Parameters
        ----------
        df
            A pandas DataFrame
        name
            The name of the pandas DataFrame
        client
            Client dictionary will be mutated with the name of the DataFrame,
            if not provided a new client is created

        Returns
        -------
        Table
            A table expression

        """
        if client is None:
            return self.connect({name: df}).table(name)
        client.dictionary[name] = df
        return client.table(name)

    def read_csv(
        self, source: str | pathlib.Path, table_name: str | None = None, **kwargs: Any
    ):
        """Register a CSV file as a table in the current session.

        Parameters
        ----------
        source
            The data source. Can be a local or remote file, pathlike objects
            also accepted.
        table_name
            An optional name to use for the created table. This defaults to
            a generated name.
        **kwargs
            Additional keyword arguments passed to Pandas loading function.
            See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_csv")
        df = pd.read_csv(source, **kwargs)
        self.dictionary[table_name] = df
        return self.table(table_name)

    def read_parquet(
        self, source: str | pathlib.Path, table_name: str | None = None, **kwargs: Any
    ):
        """Register a parquet file as a table in the current session.

        Parameters
        ----------
        source
            The data source(s). May be a path to a file, an iterable of files,
            or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a generated name.
        **kwargs
            Additional keyword arguments passed to Pandas loading function.
            See https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
            for more information.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        table_name = table_name or util.gen_name("read_parquet")
        df = pd.read_parquet(source, **kwargs)
        self.dictionary[table_name] = df
        return self.table(table_name)

    @property
    def version(self) -> str:
        return pd.__version__

    def list_tables(self, like=None, database=None):
        return self._filter_with_like(list(self.dictionary.keys()), like)

    def table(self, name: str, schema: sch.Schema | None = None):
        df = self.dictionary[name]
        schema = schema or self.schemas.get(name, None)
        schema = PandasData.infer_table(df, schema=schema)
        return ops.DatabaseTable(name, schema, self).to_expr()

    def get_schema(self, table_name, database=None):
        schemas = self.schemas
        try:
            schema = schemas[table_name]
        except KeyError:
            df = self.dictionary[table_name]
            schemas[table_name] = schema = PandasData.infer_table(df)

        return schema

    def compile(self, expr, *args, **kwargs):
        return expr

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        temp: bool | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        """Create a table."""
        if temp:
            com.IbisError(
                "Passing `temp=True` to the Pandas backend create_table method has no "
                "effect: all tables are in memory and temporary."
            )
        if database:
            com.IbisError(
                "Passing `database` to the Pandas backend create_table method has no "
                "effect: Pandas cannot set a database."
            )
        if obj is None and schema is None:
            raise com.IbisError("The schema or obj parameter is required")

        if obj is not None:
            df = self._convert_object(obj)
        else:
            dtypes = dict(PandasSchema.from_ibis(schema))
            df = pd.DataFrame(columns=dtypes.keys()).astype(dtypes)

        if name in self.dictionary and not overwrite:
            raise com.IbisError(f"Cannot overwrite existing table `{name}`")

        self.dictionary[name] = df

        if schema is not None:
            self.schemas[name] = schema
        return self.table(name)

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        return self.create_table(
            name, obj=obj, temp=None, database=database, overwrite=overwrite
        )

    def drop_view(self, name: str, *, force: bool = False) -> None:
        self.drop_table(name, force=force)

    def drop_table(self, name: str, *, force: bool = False) -> None:
        if not force and name in self.dictionary:
            raise com.IbisError(
                "Cannot drop existing table. Call drop_table with force=True to drop existing table."
            )
        del self.dictionary[name]

    def _convert_object(self, obj: Any) -> Any:
        if isinstance(obj, pd.DataFrame):
            return obj
        elif isinstance(obj, ir.Table):
            op = obj.op()
            if isinstance(op, ops.InMemoryTable):
                return op.data.to_frame()
            else:
                raise com.BackendConversionError(
                    f"Unable to convert {obj.__class__} object "
                    f"to backend type: {self.__class__.backend_table_type}"
                )
        elif isinstance(obj, pa.Table):
            return obj.to_pandas()
        else:
            raise com.BackendConversionError(
                f"Unable to convert {obj.__class__} object "
                f"to backend type: {self.__class__.backend_table_type}"
            )

    @classmethod
    @lru_cache
    def _get_operations(cls):
        from ibis.backends.pandas.kernels import supported_operations

        return supported_operations

    @classmethod
    def has_operation(cls, operation: type[ops.Value]) -> bool:
        return operation in cls._get_operations()

    def _clean_up_cached_table(self, op):
        del self.dictionary[op.name]

    def to_pyarrow(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        table_expr = expr.as_table()
        output = pa.Table.from_pandas(
            self.execute(table_expr, params=params, limit=limit, **kwargs)
        )

        # cudf.pandas adds a column with the name `__index_level_0__` (and maybe
        # other index level columns) but these aren't part of the known schema
        # so we drop them
        output = output.drop(
            filter(lambda col: col.startswith("__index_level_"), output.column_names)
        )
        table = PyArrowData.convert_table(output, table_expr.schema())
        return expr.__pyarrow_result__(table)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1000000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        pa = self._import_pyarrow()
        pa_table = self.to_pyarrow(
            expr.as_table(), params=params, limit=limit, **kwargs
        )
        return pa.RecordBatchReader.from_batches(
            pa_table.schema, pa_table.to_batches(max_chunksize=chunk_size)
        )


class Backend(BasePandasBackend):
    name = "pandas"

    def execute(self, query, params=None, limit="default", **kwargs):
        from ibis.backends.pandas.executor import PandasExecutor

        if limit != "default" and limit is not None:
            raise ValueError(
                "limit parameter to execute is not yet implemented in the "
                "pandas backend"
            )

        if not isinstance(query, ir.Expr):
            raise TypeError(
                "`query` has type {!r}, expected ibis.expr.types.Expr".format(
                    type(query).__name__
                )
            )

        params = params or {}
        params = {k.op() if isinstance(k, ir.Expr) else k: v for k, v in params.items()}

        return PandasExecutor.execute(query.op(), backend=self, params=params)

    def _load_into_cache(self, name, expr):
        self.create_table(name, expr.execute())



from __future__ import annotations

import re

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
from ibis import deferred as _
from ibis import selectors as s


@pytest.fixture
def t():
    return ibis.table(
        dict(
            a="int",
            b="string",
            c="array<string>",
            d="struct<a: array<map<string, array<float>>>>",
            e="float",
            f="decimal(3, 1)",
            g="array<array<map<float, float>>>",
            ga="string",
        ),
        name="t",
    )


@pytest.mark.parametrize(
    "sel",
    [s.where(lambda _: False), s.startswith("X"), s.endswith("🙂")],
    ids=["false", "startswith", "endswith"],
)
def test_empty_selection(t, sel):
    with pytest.raises(exc.IbisError):
        t.select(sel)


def test_where(t):
    assert t.select(s.where(lambda _: True)).equals(t.select(*t.columns))


def test_numeric(t):
    assert t.select(s.numeric()).equals(t.select("a", "e", "f"))


@pytest.mark.parametrize(
    ("obj", "expected"),
    [(dt.Array, ("c", "g")), ("float", ("e",)), (dt.Decimal(3, 1), ("f",))],
    ids=["type", "string", "instance"],
)
def test_of_type(t, obj, expected):
    assert t.select(s.of_type(obj)).equals(t.select(*expected))


@pytest.mark.parametrize(
    "name,expected",
    [
        ("array", ["c_array"]),
        ("decimal", ["c_dec52"]),
        ("floating", ["c_f32", "c_f64"]),
        ("geospatial", ["c_point"]),
        ("integer", ["c_i32", "c_u64"]),
        ("map", ["c_map"]),
        ("numeric", ["c_dec52", "c_f32", "c_f64", "c_i32", "c_u64"]),
        ("struct", ["c_struct"]),
        ("temporal", ["c_timestamp", "c_date"]),
    ],
)
def test_of_type_abstract(name, expected):
    t = ibis.table(
        dict(
            c_array="array<int>",
            c_dec52="decimal(5, 2)",
            c_f32="float32",
            c_f64="float64",
            c_point="point",
            c_i32="int32",
            c_u64="uint64",
            c_map="map<string,int>",
            c_struct="struct<a:int>",
            c_timestamp="timestamp",
            c_date="date",
        )
    )
    assert t.select(s.of_type(name)).equals(t.select(*expected))


@pytest.mark.parametrize(
    ("prefixes", "expected"),
    [("a", ("a",)), (("a", "e"), ("a", "e"))],
    ids=["string", "tuple"],
)
def test_startswith(t, prefixes, expected):
    assert t.select(s.startswith(prefixes)).equals(t.select(*expected))


def test_endswith(t):
    assert t.select(s.endswith(("a", "d"))).equals(t.select("a", "d", "ga"))


def test_contains(t):
    assert t.select(s.contains("a")).equals(t.select("a", "ga"))


@pytest.mark.parametrize(
    ("rx", "expected"),
    [("e|f", ("e", "f")), (re.compile("e|f"), ("e", "f"))],
    ids=["string", "pattern"],
)
def test_matches(t, rx, expected):
    assert t.select(s.matches(rx)).equals(t.select(expected))


def test_compose_or(t):
    assert t.select(s.contains("a") | s.startswith("d")).equals(
        t.select("a", "d", "ga")
    )


def test_compose_and(t):
    assert t.select(s.contains("a") & s.contains("g")).equals(t.select("ga"))


def test_compose_not(t):
    assert t.select(~s.numeric()).equals(t.select("b", "c", "d", "g", "ga"))


@pytest.fixture
def penguins():
    return ibis.table(
        dict(
            species="string",
            island="string",
            bill_length_mm="float64",
            bill_depth_mm="float64",
            flipper_length_mm="int64",
            body_mass_g="int64",
            sex="string",
            year="int64",
        ),
        name="penguins",
    )


def zscore(c):
    return (c - c.mean()) / c.std()


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.select(
            s.across(s.numeric() & ~s.c("year"), (_ - _.mean()) / _.std())
        ),
        lambda t: t.select(s.across(s.numeric() & ~s.c("year"), zscore)),
        lambda t: t.select(
            s.across(s.numeric() & ~s.c(t.year), (_ - _.mean()) / _.std())
        ),
        lambda t: t.select(s.across(s.numeric() & ~s.c(t.year), zscore)),
    ],
    ids=["deferred", "func", "deferred-column-ref", "func-column-ref"],
)
def test_across_select(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.select(
        bill_length_mm=zscore(_.bill_length_mm),
        bill_depth_mm=zscore(_.bill_depth_mm),
        flipper_length_mm=zscore(_.flipper_length_mm),
        body_mass_g=zscore(_.body_mass_g),
    )
    assert expr.equals(expected)


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.mutate(
            s.across(s.numeric() & ~s.c("year"), (_ - _.mean()) / _.std())
        ),
        lambda t: t.mutate(s.across(s.numeric() & ~s.c("year"), zscore)),
    ],
    ids=["deferred", "func"],
)
def test_across_mutate(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.mutate(
        bill_length_mm=zscore(_.bill_length_mm),
        bill_depth_mm=zscore(_.bill_depth_mm),
        flipper_length_mm=zscore(_.flipper_length_mm),
        body_mass_g=zscore(_.body_mass_g),
    )
    assert expr.equals(expected)


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.agg(s.across(s.numeric() & ~s.c("year"), _.mean())),
        lambda t: t.agg(s.across(s.numeric() & ~s.c("year"), lambda c: c.mean())),
    ],
    ids=["deferred", "func"],
)
def test_across_agg(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.agg(
        bill_length_mm=_.bill_length_mm.mean(),
        bill_depth_mm=_.bill_depth_mm.mean(),
        flipper_length_mm=_.flipper_length_mm.mean(),
        body_mass_g=_.body_mass_g.mean(),
    )
    assert expr.equals(expected)


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.group_by("species").select(
            s.across(s.numeric() & ~s.c("year"), (_ - _.mean()) / _.std())
        ),
        lambda t: t.group_by("species").select(
            s.across(s.numeric() & ~s.c("year"), zscore)
        ),
    ],
    ids=["deferred", "func"],
)
def test_across_group_by_select(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.group_by("species").select(
        bill_length_mm=zscore(_.bill_length_mm),
        bill_depth_mm=zscore(_.bill_depth_mm),
        flipper_length_mm=zscore(_.flipper_length_mm),
        body_mass_g=zscore(_.body_mass_g),
    )
    assert expr.equals(expected)


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.group_by("species").mutate(
            s.across(s.numeric() & ~s.c("year"), (_ - _.mean()) / _.std())
        ),
        lambda t: t.group_by("species").mutate(
            s.across(s.numeric() & ~s.c("year"), zscore)
        ),
    ],
    ids=["deferred", "func"],
)
def test_across_group_by_mutate(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.group_by("species").mutate(
        bill_length_mm=zscore(_.bill_length_mm),
        bill_depth_mm=zscore(_.bill_depth_mm),
        flipper_length_mm=zscore(_.flipper_length_mm),
        body_mass_g=zscore(_.body_mass_g),
    )
    assert expr.equals(expected)


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.group_by("species").agg(
            s.across(s.numeric() & ~s.c("year"), _.mean())
        ),
        lambda t: t.group_by("species").agg(
            s.across(s.numeric() & ~s.c("year"), lambda c: c.mean())
        ),
    ],
    ids=["deferred", "func"],
)
def test_across_group_by_agg(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.group_by("species").agg(
        bill_length_mm=_.bill_length_mm.mean(),
        bill_depth_mm=_.bill_depth_mm.mean(),
        flipper_length_mm=_.flipper_length_mm.mean(),
        body_mass_g=_.body_mass_g.mean(),
    )
    assert expr.equals(expected)


@pytest.mark.parametrize(
    "expr_func",
    [
        lambda t: t.group_by(~s.numeric()).agg(
            s.across(s.numeric() & ~s.c("year"), _.mean())
        ),
        lambda t: t.group_by(~s.numeric()).agg(
            s.across(s.numeric() & ~s.c("year"), lambda c: c.mean())
        ),
    ],
    ids=["deferred", "func"],
)
def test_across_group_by_agg_with_grouped_selectors(penguins, expr_func):
    expr = expr_func(penguins)
    expected = penguins.group_by(["species", "island", "sex"]).agg(
        bill_length_mm=_.bill_length_mm.mean(),
        bill_depth_mm=_.bill_depth_mm.mean(),
        flipper_length_mm=_.flipper_length_mm.mean(),
        body_mass_g=_.body_mass_g.mean(),
    )
    assert expr.equals(expected)


def test_across_list(penguins):
    expr = penguins.agg(s.across(["species", "island"], lambda c: c.count()))
    expected = penguins.agg(species=_.species.count(), island=_.island.count())
    assert expr.equals(expected)


def test_across_str(penguins):
    expr = penguins.agg(s.across("species", lambda c: c.count()))
    expected = penguins.agg(species=_.species.count())
    assert expr.equals(expected)


def test_if_all(penguins):
    expr = penguins.filter(s.if_all(s.numeric() & ~s.c("year"), _ > 5))
    expected = penguins.filter(
        (_.bill_length_mm > 5)
        & (_.bill_depth_mm > 5)
        & (_.flipper_length_mm > 5)
        & (_.body_mass_g > 5)
    )
    assert expr.equals(expected)


def test_if_any(penguins):
    expr = penguins.filter(s.if_any(s.numeric() & ~s.c("year"), _ > 5))
    expected = penguins.filter(
        (_.bill_length_mm > 5)
        | (_.bill_depth_mm > 5)
        | (_.flipper_length_mm > 5)
        | (_.body_mass_g > 5)
    )
    assert expr.equals(expected)


def test_negate_range(penguins):
    assert penguins.select(~s.r[3:]).equals(penguins.select(0, 1, 2))


def test_string_range_start(penguins):
    assert penguins.select(s.r["island":5]).equals(
        penguins.select(penguins.columns[penguins.columns.index("island") : 5])
    )


def test_string_range_end(penguins):
    assert penguins.select(s.r[:"island"]).equals(
        penguins.select(penguins.columns[: penguins.columns.index("island") + 1])
    )


def test_string_element(penguins):
    assert penguins.select(~s.r["island"]).equals(
        penguins.select([c for c in penguins.columns if c != "island"])
    )


def test_first(penguins):
    assert penguins.select(s.first()).equals(penguins.select(penguins.columns[0]))


def test_last(penguins):
    assert penguins.select(s.last()).equals(penguins.select(penguins.columns[-1]))


def test_all(penguins):
    assert penguins.select(s.all()).equals(penguins.select(penguins.columns))


@pytest.mark.parametrize(
    ("seq", "expected"),
    [
        param([0, 1, 2], (0, 1, 2), id="int_tuple"),
        param(~s.r[[3, 4, 5]], sorted(set(range(8)) - {3, 4, 5}), id="neg_int_list"),
        param(~s.r[3, 4, 5], sorted(set(range(8)) - {3, 4, 5}), id="neg_int_tuple"),
        param(s.r["island", "year"], ("island", "year"), id="string_tuple"),
        param(s.r[["island", "year"]], ("island", "year"), id="string_list"),
        param(iter(["island", 4, "year"]), ("island", 4, "year"), id="mixed_iterable"),
    ],
)
def test_sequence(penguins, seq, expected):
    assert penguins.select(seq).equals(penguins.select(*expected))


def test_names_callable(penguins):
    expr = penguins.select(
        s.across(
            s.numeric() & ~s.c("year"),
            func=dict(cast=_.cast("float32")),
            names=lambda col, fn: f"{fn}({col})",
        )
    )
    expected = penguins.select(
        **{
            "cast(bill_length_mm)": _.bill_length_mm.cast("float32"),
            "cast(bill_depth_mm)": _.bill_depth_mm.cast("float32"),
            "cast(flipper_length_mm)": _.flipper_length_mm.cast("float32"),
            "cast(body_mass_g)": _.body_mass_g.cast("float32"),
        }
    )
    assert expr.equals(expected)


def test_names_format_string(penguins):
    expr = penguins.select(
        s.across(
            s.numeric() & ~s.c("year"),
            func=dict(cast=_.cast("float32")),
            names="{fn}({col})",
        )
    )
    expected = penguins.select(
        **{
            "cast(bill_length_mm)": _.bill_length_mm.cast("float32"),
            "cast(bill_depth_mm)": _.bill_depth_mm.cast("float32"),
            "cast(flipper_length_mm)": _.flipper_length_mm.cast("float32"),
            "cast(body_mass_g)": _.body_mass_g.cast("float32"),
        }
    )
    assert expr.equals(expected)


def test_all_of(penguins):
    expr = penguins.select(s.all_of(s.numeric(), ~s.c("year")))
    expected = penguins.select(
        "bill_length_mm", "bill_depth_mm", "flipper_length_mm", "body_mass_g"
    )
    assert expr.equals(expected)


def test_all_of_string_list(penguins):
    # a bit silly, but robust nonetheless
    expr = penguins.select(s.all_of("year", "year"))
    expected = penguins.select("year")
    assert expr.equals(expected)


def test_any_of(penguins):
    expr = penguins.select(s.any_of(s.startswith("bill"), s.c("year")))
    expected = penguins.select("bill_length_mm", "bill_depth_mm", "year")
    assert expr.equals(expected)


def test_any_of_string_list(penguins):
    expr = penguins.select(s.any_of("year", "body_mass_g", s.matches("length")))
    expected = penguins.select(
        "bill_length_mm", "flipper_length_mm", "body_mass_g", "year"
    )
    assert expr.equals(expected)


def test_c_error_on_misspelled_column(penguins):
    match = "Columns .+ are not present"

    sel = s.c("inland")
    with pytest.raises(exc.IbisInputError, match=match):
        penguins.select(sel)

    sel = s.any_of(s.c("inland"), s.c("island"))
    with pytest.raises(exc.IbisInputError, match=match):
        penguins.select(sel)

    sel = s.any_of(s.c("island"), s.c("inland"))
    with pytest.raises(exc.IbisInputError, match=match):
        penguins.select(sel)

    sel = s.any_of(s.c("island", "inland"))
    with pytest.raises(exc.IbisInputError, match=match):
        penguins.select(sel)


def test_order_by_with_selectors(penguins):
    expr = penguins.order_by(s.of_type("string"))
    assert tuple(key.name for key in expr.op().keys) == (
        "species",
        "island",
        "sex",
    )

    expr = penguins.order_by(s.all())
    assert tuple(key.name for key in expr.op().keys) == tuple(expr.columns)

    with pytest.raises(exc.IbisError):
        penguins.order_by(~s.all())



"""Test utilities."""

from __future__ import annotations

import pickle
from typing import Callable

import ibis
import ibis.expr.types as ir
from ibis import util


def assert_equal(left, right):
    """Assert that two ibis objects are equal."""

    if util.all_of([left, right], ibis.Schema):
        assert left.equals(right), f"Comparing schemas: \n{left!r} !=\n{right!r}"
    else:
        assert left.equals(right), f"Objects unequal: \n{left!r}\nvs\n{right!r}"


def assert_pickle_roundtrip(obj):
    """Assert that an ibis object remains the same after pickling and
    unpickling."""
    loaded = pickle.loads(pickle.dumps(obj))
    if hasattr(obj, "equals"):
        assert obj.equals(loaded)
    else:
        assert obj == loaded


def schemas_eq(left: ir.Expr, right: ir.Expr) -> bool:
    left_schema = left.as_table().schema()
    right_schema = right.as_table().schema()
    return left_schema == right_schema


def assert_decompile_roundtrip(
    expr: ir.Expr,
    snapshot=None,
    eq: Callable[[ir.Expr, ir.Expr], bool] = ir.Expr.equals,
):
    """Assert that an ibis expression remains the same after decompilation.

    Parameters
    ----------
    expr
        The expression to decompile.
    snapshot
        A snapshot fixture.
    eq
        A callable that returns whether two Ibis expressions are equal.
        Defaults to `ibis.expr.types.Expr.equals`. Use this to adjust
        comparison behavior for expressions that contain `SelfReference`
        operations from table.view() calls, or other relations whose equality
        is difficult to roundtrip.
    """
    rendered = ibis.decompile(expr, format=True)
    if snapshot is not None:
        snapshot.assert_match(rendered, "decompiled.py")

    # execute the rendered python code
    locals_ = {}
    exec(rendered, {}, locals_)
    restored = locals_["result"]

    assert eq(expr.unbind(), restored)



from __future__ import annotations

import math
import string

import pytest

import ibis
import ibis.common.exceptions as exc
from ibis import udf, util
from ibis.backends.trino.tests.conftest import (
    TRINO_HOST,
    TRINO_PASS,
    TRINO_PORT,
    TRINO_USER,
)


@pytest.fixture
def tmp_name(con):
    name = ibis.util.gen_name("test_trino")
    yield name
    con.drop_table(name, force=True)


def test_table_properties(tmp_name):
    con = ibis.trino.connect(database="hive", schema="default")
    schema = ibis.schema(dict(a="int"))
    t = con.create_table(
        tmp_name,
        schema=schema,
        properties={
            "format": "ORC",
            "bucketed_by": ["a"],
            "bucket_count": 42,
            "extra_properties": {
                "any": "property",
                "you": "want",
            },
        },
    )
    assert t.schema() == schema
    with con.begin() as c:
        c.execute(f"SHOW CREATE TABLE {tmp_name}")
        [(ddl,)] = c.fetchall()
    assert "ORC" in ddl
    assert "bucketed_by" in ddl


def test_hive_table_overwrite(tmp_name):
    con = ibis.trino.connect(database="hive", schema="default")
    schema = ibis.schema(dict(a="int"))

    t = con.create_table(tmp_name, schema=schema)
    assert tmp_name in con.list_tables()
    assert t.schema() == schema

    t = con.create_table(tmp_name, schema=schema, overwrite=True)
    assert tmp_name in con.list_tables()
    assert t.schema() == schema


def test_list_catalogs(con):
    assert {"hive", "memory", "system", "tpch", "tpcds"}.issubset(con.list_databases())


def test_list_schemas(con):
    assert {"information_schema", "sf1"}.issubset(con.list_schemas(database="tpch"))


@pytest.mark.parametrize(("source", "expected"), [(None, "ibis"), ("foo", "foo")])
def test_con_source(source, expected):
    con = ibis.trino.connect(
        user=TRINO_USER,
        host=TRINO_HOST,
        port=TRINO_PORT,
        password=TRINO_PASS,
        database="hive",
        schema="default",
        source=source,
    )
    assert con.con.source == expected


@pytest.mark.parametrize(
    ("database", "schema", "table"),
    [
        # tables known to exist
        ("system", "metadata", "table_comments"),
        ("tpcds", "sf1", "store"),
        ("tpch", "sf1", "nation"),
    ],
)
def test_cross_schema_table_access(con, database, schema, table):
    t = con.table(table, schema=schema, database=database)
    assert t.count().execute()


def test_builtin_scalar_udf(con, snapshot):
    @udf.scalar.builtin
    def bar(x: float, width: int) -> str:
        """Render a single bar of length `width`, with `x` percent filled."""

    expr = bar(0.25, 40)
    result = con.execute(expr)
    snapshot.assert_match(result, "result.txt")


def test_builtin_agg_udf(con):
    @udf.agg.builtin
    def geometric_mean(x) -> float:
        """Geometric mean of a series of numbers."""

    t = con.table("diamonds")
    expr = t.agg(n=t.count(), geomean=geometric_mean(t.price))
    result_n, result = expr.execute().squeeze().tolist()

    with con.begin() as c:
        c.execute("SELECT COUNT(*), GEOMETRIC_MEAN(price) FROM diamonds")
        [(expected_n, expected)] = c.fetchall()

    # check the count
    assert result_n > 0
    assert expected_n > 0
    assert result_n == expected_n

    # check the value
    assert result is not None
    assert expected is not None
    assert math.isfinite(result)
    assert result == expected


def test_create_table_timestamp():
    con = ibis.trino.connect(database="memory", schema="default")
    schema = ibis.schema(
        dict(zip(string.ascii_letters, map("timestamp({:d})".format, range(10))))
    )
    table = util.gen_name("trino_temp_table")
    t = con.create_table(table, schema=schema)
    try:
        rows = con.raw_sql(f"DESCRIBE {table}").fetchall()
        result = ibis.schema((name, typ) for name, typ, *_ in rows)
        assert result == schema
        assert result == t.schema()
    finally:
        con.drop_table(table)
        assert table not in con.list_tables()


def test_table_access_database_schema(con):
    t = con.table("region", schema="sf1", database="tpch")
    assert t.count().execute()

    with pytest.raises(exc.IbisError, match='Table not found: tpch."tpch.sf1".region'):
        con.table("region", schema="tpch.sf1", database="tpch")

    with pytest.raises(
        exc.IbisError, match='Table not found: system."tpch.sf1".region'
    ):
        con.table("region", schema="tpch.sf1", database="system")
